// Global objects/variables
//-----------------------------------------
SamplerState SamPointClamp
{
    Filter = MIN_MAG_MIP_POINT;
    AddressU = Clamp;
    AddressV = Clamp;
    AddressW = Clamp;
};

SamplerState SamLinear
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
    AddressW = Clamp;
};

RasterizerState CullNone
{
    CullMode = None;
};

BlendState AlphaBlending
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = TRUE;
    SrcBlend = SRC_ALPHA;
    DestBlend = INV_SRC_ALPHA;
    BlendOp = ADD;
    SrcBlendAlpha = ONE;
    DestBlendAlpha = ONE;
    BlendOpAlpha = ADD;
    RenderTargetWriteMask[0] = 0x0F;
}; 

BlendState AlphaBlendingColor
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = TRUE;
    SrcBlend = SRC_COLOR;
    DestBlend = INV_SRC_COLOR;
    BlendOp = ADD;
    SrcBlendAlpha = ONE;
    DestBlendAlpha = ONE;
    BlendOpAlpha = ADD;
    RenderTargetWriteMask[0] = 0x0F;
}; 

BlendState NoAlphaBlending
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = FALSE;
    RenderTargetWriteMask[0] = 0x0F;
};

// Textures
Texture3D TextureVelocity0;
Texture3D TextureVelocity1;
Texture3D TextureVelocityOld;
Texture3D TextureColor0;
Texture3D TextureColor1;
Texture3D TextureTempScalar0;
Texture3D TextureTempScalar1;
Texture3D TexturePressure;
Texture2D TextureBase;

// Variables
float TextureWidth, TextureHeight, TextureDepth;
float Timestep;
float3 Color, Color1, Color2;
float4 Sphere;


// Helper Functions
//-----------------------------------------

float3 ToTexCoord(float3 x)
{
	return float3(x.x / TextureWidth, x.y / TextureHeight, (x.z + 0.5f) / TextureDepth);
}

// Global Vertex/GeometryShader
//-----------------------------------------

struct VS_OUTPUT
{
    float4 Pos				 : SV_POSITION;
    float3 Tex				 : TEXCOORD0;	// 3D cell texcoords
    float3 TexScaled		 : TEXCOORD1;
};

struct GS_OUTPUT
{
    float4 Pos               : SV_Position;
    float3 Tex				 : TEXCOORD0;
    float3 TexScaled		 : TEXCOORD1;
    uint RTIndex             : SV_RenderTargetArrayIndex;  // Destination slice
};

VS_OUTPUT VS(float4 pos : POSITION)
{
    VS_OUTPUT output;
    output.Pos.xy = pos.xy;
    output.Pos.z = 0;
    output.Pos.w = 1;
    
    output.TexScaled.x = pos.x / 2.0f + 0.5f;
    output.TexScaled.y = 1.0f - (pos.y / 2.0f + 0.5f);
    output.TexScaled.z = pos.z + 0.5f / TextureDepth;
    
    output.Tex.x = (int)(output.TexScaled.x * TextureWidth);
	output.Tex.y = (int)(output.TexScaled.y * TextureHeight);
    output.Tex.z = (int)(pos.z * TextureDepth);
    
    return output;
}

[maxvertexcount(3)]
void GS(triangle VS_OUTPUT input[3], inout TriangleStream<GS_OUTPUT> triStream)
{
    GS_OUTPUT output;
    
    output.RTIndex = input[0].Tex.z;
    
    for(int v = 0; v < 3; v++)
    {
        output.Pos = input[v].Pos; 
        output.Tex = input[v].Tex;
        output.TexScaled = input[v].TexScaled;
        triStream.Append(output);
    }
    
    triStream.RestartStrip();
}

// Velocity advection
//-----------------------------------------

float4 PSAdvectVelocity(GS_OUTPUT input) : SV_Target
{	
	float3 pos = input.Tex;
	
    pos -= Timestep * TextureVelocity1.SampleLevel(SamPointClamp, input.TexScaled, 0).xyz;

	float4 vel = TextureVelocity1.SampleLevel(SamLinear, ToTexCoord(pos), 0);
	
	vel -= vel * 0.5f * Timestep;
	return vel;
}

// TextureVelocity1 = phi^
float4 PSAdvectVelocityMc(GS_OUTPUT input) : SV_Target
{	
    float3 advPos = input.Tex - Timestep * TextureVelocityOld.SampleLevel(SamPointClamp, input.TexScaled, 0 ).xyz;
    
    float3 advPosScaled = ToTexCoord(advPos);

    float3 advPosCenter = floor(advPos + float3(0.5f, 0.5f, 0.5f));
    float3 advPosCenterScaled = float3(advPosCenter.x / TextureWidth, advPosCenter.y / TextureHeight, advPosCenter.z / TextureDepth);

    float3 ht = float3(0.5f / TextureWidth, 0.5f / TextureHeight, 0.5f / TextureDepth);

    float4 nodeValues[8];
    nodeValues[0] = TextureVelocityOld.SampleLevel(SamPointClamp, advPosCenterScaled + float3(-ht.x, -ht.y, -ht.z), 0);
    nodeValues[1] = TextureVelocityOld.SampleLevel(SamPointClamp, advPosCenterScaled + float3(-ht.x, -ht.y,  ht.z), 0);
    nodeValues[2] = TextureVelocityOld.SampleLevel(SamPointClamp, advPosCenterScaled + float3(-ht.x,  ht.y, -ht.z), 0);
    nodeValues[3] = TextureVelocityOld.SampleLevel(SamPointClamp, advPosCenterScaled + float3(-ht.x,  ht.y,  ht.z), 0);
    nodeValues[4] = TextureVelocityOld.SampleLevel(SamPointClamp, advPosCenterScaled + float3( ht.x, -ht.y, -ht.z), 0);
    nodeValues[5] = TextureVelocityOld.SampleLevel(SamPointClamp, advPosCenterScaled + float3( ht.x, -ht.y,  ht.z), 0);
    nodeValues[6] = TextureVelocityOld.SampleLevel(SamPointClamp, advPosCenterScaled + float3( ht.x,  ht.y, -ht.z), 0);
    nodeValues[7] = TextureVelocityOld.SampleLevel(SamPointClamp, advPosCenterScaled + float3( ht.x,  ht.y,  ht.z), 0);

    float4 phiMin = min(min(min(nodeValues[0], nodeValues [1]), nodeValues [2]), nodeValues [3]);
    phiMin = min(min(min(min(phiMin, nodeValues [4]), nodeValues [5]), nodeValues [6]), nodeValues [7]);
    
    float4 phiMax = max(max(max(nodeValues[0], nodeValues [1]), nodeValues [2]), nodeValues [3]);
    phiMax = max(max(max(max(phiMax, nodeValues [4]), nodeValues [5]), nodeValues [6]), nodeValues [7]);

    float4 result = TextureVelocityOld.SampleLevel(SamLinear, advPosScaled, 0) + 0.5 * (TextureVelocityOld.SampleLevel(SamPointClamp, input.TexScaled, 0) - TextureVelocity1.SampleLevel(SamPointClamp, input.TexScaled, 0));

    result = max(min(result, phiMax), phiMin);
    
    result -= result * 0.5f * Timestep;

    return result;
}

// Color advection
//-----------------------------------------

float2 PSAdvectColor(GS_OUTPUT input) : SV_Target
{
    float3 pos = input.TexScaled;
	
	pos = input.Tex - Timestep * TextureVelocity0.SampleLevel(SamPointClamp, pos, 0).xyz;
	
	float2 color = TextureColor1.SampleLevel(SamLinear, ToTexCoord(pos), 0);
	
	color.x -= color.x * Timestep * 2.5f;
    color.y -= color.y * Timestep * 1.2f;
	
	return color;
}

// TextureVelocity1 = phi^
float4 PSAdvectColorMc(GS_OUTPUT input) : SV_Target
{	
    float3 advPos = input.Tex - Timestep * TextureVelocityOld.SampleLevel(SamPointClamp, input.TexScaled, 0 ).xyz;
    
    float3 advPosScaled = ToTexCoord(advPos);

    float3 advPosCenter = floor(advPos + float3(0.5f, 0.5f, 0.5f));
    float3 advPosCenterScaled = float3(advPosCenter.x / TextureWidth, advPosCenter.y / TextureHeight, advPosCenter.z / TextureDepth);

    float3 ht = float3(0.5f / TextureWidth, 0.5f / TextureHeight, 0.5f / TextureDepth);

    float4 nodeValues[8];
    nodeValues[0] = TextureColor1.SampleLevel(SamPointClamp, advPosCenterScaled + float3(-ht.x, -ht.y, -ht.z), 0);
    nodeValues[1] = TextureColor1.SampleLevel(SamPointClamp, advPosCenterScaled + float3(-ht.x, -ht.y,  ht.z), 0);
    nodeValues[2] = TextureColor1.SampleLevel(SamPointClamp, advPosCenterScaled + float3(-ht.x,  ht.y, -ht.z), 0);
    nodeValues[3] = TextureColor1.SampleLevel(SamPointClamp, advPosCenterScaled + float3(-ht.x,  ht.y,  ht.z), 0);
    nodeValues[4] = TextureColor1.SampleLevel(SamPointClamp, advPosCenterScaled + float3( ht.x, -ht.y, -ht.z), 0);
    nodeValues[5] = TextureColor1.SampleLevel(SamPointClamp, advPosCenterScaled + float3( ht.x, -ht.y,  ht.z), 0);
    nodeValues[6] = TextureColor1.SampleLevel(SamPointClamp, advPosCenterScaled + float3( ht.x,  ht.y, -ht.z), 0);
    nodeValues[7] = TextureColor1.SampleLevel(SamPointClamp, advPosCenterScaled + float3( ht.x,  ht.y,  ht.z), 0);

    float4 phiMin = min(min(min(nodeValues[0], nodeValues [1]), nodeValues [2]), nodeValues [3]);
    phiMin = min(min(min(min(phiMin, nodeValues [4]), nodeValues [5]), nodeValues [6]), nodeValues [7]);
    
    float4 phiMax = max(max(max(nodeValues[0], nodeValues [1]), nodeValues [2]), nodeValues [3]);
    phiMax = max(max(max(max(phiMax, nodeValues [4]), nodeValues [5]), nodeValues [6]), nodeValues [7]);

    float4 result = TextureColor1.SampleLevel(SamLinear, advPosScaled, 0) + 0.5 * (TextureColor1.SampleLevel(SamPointClamp, input.TexScaled, 0) - TextureVelocity1.SampleLevel(SamPointClamp, input.TexScaled, 0));

    result = max(min(result, phiMax), phiMin);
    
    result.x -= result.x * Timestep * 2.5f;
    result.y -= result.y * Timestep * 1.2f;

    return result;
}

// Divergence computation
//-----------------------------------------

float PSComputeDivergence(GS_OUTPUT input) : SV_Target
{    
    float3 pos = input.TexScaled;
	
	float3 center = TextureVelocity1.SampleLevel(SamPointClamp, pos, 0);

	float3 leftAddr   = ToTexCoord(float3(input.Tex.x - 1, input.Tex.y, input.Tex.z));
	float3 rightAddr  = ToTexCoord(float3(input.Tex.x + 1, input.Tex.y, input.Tex.z));
	float3 bottomAddr = ToTexCoord(float3(input.Tex.x, input.Tex.y - 1, input.Tex.z));
	float3 topAddr    = ToTexCoord(float3(input.Tex.x, input.Tex.y + 1, input.Tex.z));
	float3 downAddr   = ToTexCoord(float3(input.Tex.x, input.Tex.y, input.Tex.z - 1));
	float3 upAddr     = ToTexCoord(float3(input.Tex.x, input.Tex.y, input.Tex.z + 1));
	
	float3 left   = TextureVelocity1.SampleLevel(SamPointClamp, leftAddr, 0);
    float3 right  = TextureVelocity1.SampleLevel(SamPointClamp, rightAddr, 0);
    float3 bottom = TextureVelocity1.SampleLevel(SamPointClamp, bottomAddr, 0);
    float3 top    = TextureVelocity1.SampleLevel(SamPointClamp, topAddr, 0);
    float3 down   = TextureVelocity1.SampleLevel(SamPointClamp, downAddr, 0);
    float3 up     = TextureVelocity1.SampleLevel(SamPointClamp, upAddr, 0);

    return 0.5 * (right.x - left.x + top.y - bottom.y + up.z - down.z);
}

// Jacobi computation
//-----------------------------------------

float PSComputeJacobi(GS_OUTPUT input) : SV_Target
{
    float3 pos = input.TexScaled;

	float center = TexturePressure.SampleLevel(SamPointClamp, pos, 0);
    
    float div = TextureTempScalar0.SampleLevel(SamPointClamp, pos, 0);
    
    float3 leftAddr   = ToTexCoord(float3(input.Tex.x - 1, input.Tex.y, input.Tex.z));
	float3 rightAddr  = ToTexCoord(float3(input.Tex.x + 1, input.Tex.y, input.Tex.z));
	float3 bottomAddr = ToTexCoord(float3(input.Tex.x, input.Tex.y - 1, input.Tex.z));
	float3 topAddr    = ToTexCoord(float3(input.Tex.x, input.Tex.y + 1, input.Tex.z));
	float3 downAddr   = ToTexCoord(float3(input.Tex.x, input.Tex.y, input.Tex.z - 1));
	float3 upAddr     = ToTexCoord(float3(input.Tex.x, input.Tex.y, input.Tex.z + 1));
	
	float left   = TexturePressure.SampleLevel(SamPointClamp, leftAddr, 0);
    float right  = TexturePressure.SampleLevel(SamPointClamp, rightAddr, 0);
    float bottom = TexturePressure.SampleLevel(SamPointClamp, bottomAddr, 0);
    float top    = TexturePressure.SampleLevel(SamPointClamp, topAddr, 0);
    float down   = TexturePressure.SampleLevel(SamPointClamp, downAddr, 0);
    float up     = TexturePressure.SampleLevel(SamPointClamp, upAddr, 0);

    return (left + right + bottom + top + up + down - div) / 6.0f;
}

// Projection
//-----------------------------------------

float3 PSProject(GS_OUTPUT input) : SV_Target
{  
	float3 pos = input.TexScaled;
	
	float center = TexturePressure.SampleLevel(SamPointClamp, pos, 0);

    float3 leftAddr   = ToTexCoord(float3(input.Tex.x - 1, input.Tex.y, input.Tex.z));
	float3 rightAddr  = ToTexCoord(float3(input.Tex.x + 1, input.Tex.y, input.Tex.z));
	float3 bottomAddr = ToTexCoord(float3(input.Tex.x, input.Tex.y - 1, input.Tex.z));
	float3 topAddr    = ToTexCoord(float3(input.Tex.x, input.Tex.y + 1, input.Tex.z));
	float3 downAddr   = ToTexCoord(float3(input.Tex.x, input.Tex.y, input.Tex.z - 1));
	float3 upAddr     = ToTexCoord(float3(input.Tex.x, input.Tex.y, input.Tex.z + 1));
	
	float left   = TexturePressure.SampleLevel(SamPointClamp, leftAddr, 0);
    float right  = TexturePressure.SampleLevel(SamPointClamp, rightAddr, 0);
    float bottom = TexturePressure.SampleLevel(SamPointClamp, bottomAddr, 0);
    float top    = TexturePressure.SampleLevel(SamPointClamp, topAddr, 0);
    float down   = TexturePressure.SampleLevel(SamPointClamp, downAddr, 0);
    float up     = TexturePressure.SampleLevel(SamPointClamp, upAddr, 0);

    return (TextureVelocity1.SampleLevel(SamPointClamp, pos, 0).xyz - (0.5 * float3(right - left, top - bottom, up - down)));
}

float2 PSFillSphere(GS_OUTPUT input) : SV_Target
{
    float3 coord = input.TexScaled;
    float dist = length(Sphere.xyz - coord) - Sphere.w;
    if(dist < 0)
		return Color.xy;
	else
		return TextureColor0.SampleLevel(SamPointClamp, coord, 0);
}

float4 PSFillSphereVelocity(GS_OUTPUT input) : SV_Target
{
	float3 coord = input.TexScaled;
    float dist = length(Sphere.xyz - coord) - Sphere.w;
    if(dist < 0)
		return float4(Color, dist);
	else
		return TextureVelocity0.SampleLevel(SamPointClamp, coord, 0);
}

float4 PSDrawBase(GS_OUTPUT input) : SV_Target
{
	if(input.Tex.y < 1)
	{
		float3 texVal = TextureBase.SampleLevel(SamPointClamp, input.TexScaled.xz, 0);
		texVal.r = (texVal.r + texVal.g + texVal.b) / 3;
		return float4(Color * texVal.r, 1);
	}
	else
		return TextureColor0.SampleLevel(SamPointClamp, input.TexScaled, 0);
}

float4 PSDrawBaseVelocity(GS_OUTPUT input) : SV_Target
{
	if(input.Tex.y < 1)
	{
		float3 finalColor;
		float3 texVal = TextureBase.SampleLevel(SamPointClamp, input.TexScaled.xz, 0);
		if(texVal.r > texVal.g && texVal.r > texVal.b)
			finalColor = Color;
		else
			if(texVal.g > texVal.r && texVal.g > texVal.b)
				finalColor = Color1;
			else
				finalColor = Color2;
		
		texVal = texVal * 2 - 1;
		return float4(finalColor.r * texVal.r, finalColor.g * abs(texVal.g), finalColor.b * texVal.b, 1);
	}
	else
		return TextureColor0.SampleLevel(SamPointClamp, input.TexScaled, 0);
}

//-----------------------------------------

technique10 AdvectVelocity
{
    pass P0
    {
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(CompileShader(gs_4_0, GS()));
        SetPixelShader(CompileShader(ps_4_0, PSAdvectVelocity()));
        SetRasterizerState(CullNone);
		SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 AdvectVelocityMc
{
    pass P0
    {
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(CompileShader(gs_4_0, GS()));
        SetPixelShader(CompileShader(ps_4_0, PSAdvectVelocityMc()));
        SetRasterizerState(CullNone);
		SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 AdvectColor
{
    pass P0
    {
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(CompileShader(gs_4_0, GS()));
        SetPixelShader(CompileShader(ps_4_0, PSAdvectColor()));
        SetRasterizerState(CullNone);
		SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 AdvectColorMc
{
    pass P0
    {
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(CompileShader(gs_4_0, GS()));
        SetPixelShader(CompileShader(ps_4_0, PSAdvectColorMc()));
        SetRasterizerState(CullNone);
		SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 ComputeDivergence
{
    pass P0
    {
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(CompileShader(gs_4_0, GS()));
        SetPixelShader(CompileShader(ps_4_0, PSComputeDivergence()));
        SetRasterizerState(CullNone);
		SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 ComputeJacobi
{
    pass P0
    {
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(CompileShader(gs_4_0, GS()));
        SetPixelShader(CompileShader(ps_4_0, PSComputeJacobi()));
        SetRasterizerState(CullNone);
		SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 Project
{
    pass P0
    {
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(CompileShader(gs_4_0, GS()));
        SetPixelShader(CompileShader(ps_4_0, PSProject()));
        SetRasterizerState(CullNone);
		SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 FillSphere
{
    pass P0
    {
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(CompileShader(gs_4_0, GS()));
        SetPixelShader(CompileShader(ps_4_0, PSFillSphere()));
        SetRasterizerState(CullNone);
        SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 FillSphereVelocity
{
    pass P0
    {
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(CompileShader(gs_4_0, GS()));
        SetPixelShader(CompileShader(ps_4_0, PSFillSphereVelocity()));
        SetRasterizerState(CullNone);
        SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 DrawBase
{
    pass P0
    {
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(CompileShader(gs_4_0, GS()));
        SetPixelShader(CompileShader(ps_4_0, PSDrawBase()));
        SetRasterizerState(CullNone);
        SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 DrawBaseVelocity
{
    pass P0
    {
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(CompileShader(gs_4_0, GS()));
        SetPixelShader(CompileShader(ps_4_0, PSDrawBaseVelocity()));
        SetRasterizerState(CullNone);
        SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}