matrix World;
matrix View;
matrix Projection;
matrix WorldViewProjection;

float3 LightDir[3];
float3 LightColor[3];
float3 AmbientColor;
float3 DiffuseColor;
float3 SpecularColor;
float3 EmissiveColor;
float3 CameraPos;
float Shininess;
int NumLights;
float Opacity;
float GaussianNoise[5];

float3 DMParameters;

Texture2DArray WaveTexture;
Texture2D ReflectionTexture;
Texture2D RefractionTexture;
matrix ReflectionMatrix;
int WaveTextureSlide1;
int WaveTextureSlide2;

float WaveTextureOffset;

RasterizerState DisableCulling
{
    CullMode = NONE;
};

RasterizerState WireFrame
{
	FillMode = WIREFRAME;
	CullMode = NONE;
};

BlendState NoAlpha
{
    BlendEnable[0] = FALSE;
};

//--------------------------------------------------------------------------------------
struct VS_INPUT
{
    float4 Pos : POSITION;
};

struct PS_INPUT
{
    float4 Pos : SV_POSITION;
    float4 ProjCoord: TEXCOORD1;
	//float3 PosInWorld : TEXCOORD2; 
	float3 EyeVec : TEXCOORD2;
	float2 TexCoord: TEXCOORD3;
};

SamplerState SamLinear
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Wrap;
    AddressV = Wrap;
};

SamplerState SamLinearClamp
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
};

SamplerState SamAnisotropic
{
    AddressU = Wrap;
    AddressV = Wrap;
    Filter = ANISOTROPIC;
    MaxAnisotropy = 16;
};
//--------------------------------------------------------------------------------------

PS_INPUT VS(VS_INPUT input)
{
    PS_INPUT output;
    input.Pos.xz = input.Pos.xz + CameraPos.xz;      
    input.Pos.y = DMParameters.y;
    float4 ClipPos = mul(input.Pos, WorldViewProjection);
    float3 b0 = abs(ClipPos.xyz) < (ClipPos.www * 2. + 20);
    float4 texVal = 0.0;
    output.TexCoord = input.Pos.xz * DMParameters.x;
    
    //No vertex displacement when vertex point is not in viewport. 
    if (b0.x > 0 && b0.y > 0 && b0.z > 0)
    {
		float dist = length(abs(CameraPos - input.Pos));
		if (dist <= 5000)
		{
			//No vertex displacement at a certain distance,
			//for performance reason
			float4 texVal1 = (float4)WaveTexture.SampleLevel(SamAnisotropic, float3(output.TexCoord.x + WaveTextureOffset, output.TexCoord.y, WaveTextureSlide1),0);	
			float4 texVal2 = (float4)WaveTexture.SampleLevel(SamAnisotropic, float3(output.TexCoord.x + WaveTextureOffset, output.TexCoord.y, WaveTextureSlide2),0);	
			texVal = lerp (texVal1, texVal2, WaveTextureOffset);
			input.Pos.y = DMParameters.y + (texVal.a - 0.5) * DMParameters.z ;	
		}
	}	
	output.EyeVec = CameraPos - input.Pos;
    output.Pos = mul(input.Pos, WorldViewProjection);
    output.ProjCoord = mul( input.Pos, mul(WorldViewProjection,ReflectionMatrix));         
    
    return output;
}

float3 BumpGen( PS_INPUT input,float3 norm  )
{
  float3 bumpNormal;

  // first layer bump, for big waves
  float3 bumpColorA = WaveTexture.Sample(SamAnisotropic, float3(input.TexCoord * 4,WaveTextureSlide1) ); 
  float3 bumpColorB = WaveTexture.Sample(SamAnisotropic, float3(input.TexCoord * 4 + WaveTextureOffset*4, WaveTextureSlide1));         

  float3 bumpLowFreq;
  bumpLowFreq = ((bumpColorA + bumpColorB) - 1.0) * 2;
              
  //Second layer bump, small waves             
  float3 bumpColorC = WaveTexture.Sample(SamAnisotropic, float3(input.TexCoord * 8,WaveTextureSlide1));           
  float3 bumpColorD = WaveTexture.Sample(SamAnisotropic, float3(input.TexCoord * 8 + WaveTextureOffset*8,WaveTextureSlide1));       
  
  float3 bumpHighFreq;
  
  bumpHighFreq = (bumpColorC + bumpColorD) - 1.0;    

  // Mix high and low frequency normals
  bumpNormal = float3(bumpLowFreq.xzy + bumpHighFreq.xzy);

  // Clamp max bump to 0.15
  bumpNormal.xz *= 0.15;

  bumpNormal.y = 1;
  bumpNormal.xyz += norm.xyz;

  //bumpNormal.xyz = normalize( bumpNormal );

  return bumpNormal;
}

float4 PS(PS_INPUT input) : SV_Target
{
	//Todo (make parameter)
	//fresnelValue.xyz = power,scale, bias
	float3 fresnelValues = float3(5.0, 0.97963, 0.02037);	
	float3 norm1 = (float3)WaveTexture.Sample(SamAnisotropic, float3(input.TexCoord.x + WaveTextureOffset, input.TexCoord.y, WaveTextureSlide1));
	float3 norm2 = (float3)WaveTexture.Sample(SamAnisotropic, float3(input.TexCoord.x + WaveTextureOffset, input.TexCoord.y, WaveTextureSlide2));
	float3 norm = lerp (norm1, norm2, WaveTextureOffset);
	norm = normalize(norm.xzy * 2.0 - 1.0);
	
 	float3 EyeVec = normalize(input.EyeVec);
    float2 TexCoords = input.ProjCoord.xy / input.ProjCoord.w;
   
	// Read additional normal vectors from texture 
	float3 bumpNormal = BumpGen( input, norm );
	//bumpNormal.y = 1;
	float3 finalNormal = normalize( norm.xyz + bumpNormal );
	
    float3 SpecularColor = 0;
	
	float NdotI = max(dot( EyeVec.xyz, finalNormal.xyz ), 0);

	//Calculate specular color	
	for (int i=0; i < NumLights; i++)
	{
		float3 halfAngle = normalize(normalize(CameraPos) - LightDir[i]);
		float nDotH = clamp(dot(halfAngle, norm),0.0,1.0);		
		SpecularColor += pow(nDotH, Shininess);
	}
	
	//Calculate Fresnel factor
	float Facing = 1.0 - NdotI;
    float fresnelTerm = fresnelValues.z + pow(Facing, fresnelValues.x) * fresnelValues.y;
        
   //Ambient color is deep water color, diffuse color is shallow water color
   float3 WaterColor = lerp(AmbientColor, DiffuseColor, Facing);
  
   float3 RefrColor = RefractionTexture.Sample(SamLinearClamp, TexCoords + finalNormal.xz * 0.10 * 0.5 );
   RefrColor.rgb = lerp(RefrColor.rgb, WaterColor, Opacity);

   //Do 5 samples of reflection with a slight distortion       
   float3 ReflColor = 0;
   for (int i = 0; i < 5; i++)
	  ReflColor += ReflectionTexture.Sample(SamLinearClamp, (TexCoords + finalNormal.xz * 0.10) + GaussianNoise[i] * 0.005) * 0.2;

   float3 FinalColor = lerp(RefrColor.rgb, ReflColor.rgb, min(0.95,fresnelTerm + SpecularColor*0.5));
   //FinalColor = SpecularColor;
   return float4(FinalColor.rgb, 1.0);
}

//--------------------------------------------------------------------------------------
technique10 Render
{
    pass P0
    {
		//SetRasterizerState(DisableCulling);   
		//SetRasterizerState(WireFrame);
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PS()));
        SetBlendState(NoAlpha, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}
