matrix World;
matrix View;
matrix Projection;
matrix WorldViewProjection;

float2 SourceSize;           

// Clearing to constant
float4 ConstValue = 0;

// Blurring
float2 BlurDimension;
int    BlurSamples;


//Gaussian Blur
#define MAX_GAUSSIANSAMPLES 17
float	GaussianBlurWeights[MAX_GAUSSIANSAMPLES];
int		GaussianBlurSamples;
float2  GaussianBlurDimension;

typedef Texture2D<uint4> UintTex;

Texture2D SourceTexture;
Texture2DArray SourceTextureArray;
UintTex SourceTextureUint;

Texture2D Luminance1x1Texture;
Texture2D BlurTexture;


float TimeSinceLastFrame;
float Exposure;
float BlurWeight;
float LuminanceMaxAdaption = 0.1f;		//smaller value = more adaption range

float BlendFactor;

static const float3 LUMINANCE_WEIGHTS = float3(.299f, .587f, .114f);
static const float ADAPTION_SPEED = 30.f;			//How fast eye adapts to new luminance, smaller values = slower adaption
static const float  BRIGHT_THRESHOLD = 0.35f;

//--------------------------------------------------------------------------------------
// State
//--------------------------------------------------------------------------------------
RasterizerState rsPostProcess
{
    CullMode = None;
    FillMode = Solid;
    ScissorEnable = true;
    MultisampleEnable = false;
};

RasterizerState rsPostProcessNoScissor
{
    CullMode = None;
    FillMode = Solid;
    ScissorEnable = false;
    MultisampleEnable = false;
};

shared DepthStencilState dsPostProcess
{
    DepthEnable = false;
    DepthWriteMask = 0;
    StencilEnable = false;
};

BlendState bsNormal
{
};

BlendState SrcAlphaBlendingAdd
{
        BlendEnable[0] = TRUE;
        SrcBlend = SRC_ALPHA;
        DestBlend = INV_SRC_ALPHA;
        BlendOp = ADD;
        SrcBlendAlpha = ZERO;
        DestBlendAlpha = ONE;
        BlendOpAlpha = ADD;
        RenderTargetWriteMask[0] = 0x0F;	//All colors
};

// Basic bilinear sampling
SamplerState sampBilinear
{
    AddressU = Clamp;
    AddressV = Clamp;
    Filter = MIN_MAG_LINEAR_MIP_POINT;
};

SamplerState sampPoint
{
    AddressU = Clamp;
    AddressV = Clamp;
    Filter = MIN_MAG_MIP_POINT;
};

struct VSIn
{
    uint VertexID   : SV_VertexID;
};

struct PSIn
{
    float4 Position : SV_Position;
    float2 TexCoord : TEXCOORD;
};

struct PSInToneMapping
{
    float4 Pos    : SV_POSITION;
    float3 Tex    : TEXCOORD0;
};

//-------------------------------------------------------------
// Generate a full-screen triangle from vertex ID's
PSIn PostProcess_VS(VSIn Input)
{
    PSIn Output;
    
    Output.TexCoord = float2((Input.VertexID << 1) & 2, Input.VertexID & 2);
    Output.Position = float4(Output.TexCoord * float2(2, -2) + float2(-1, 1), 0, 1);
    
    return Output;
}

PSInToneMapping ToneMapping_VS(VSIn Input)
{
    PSInToneMapping Output;
    
    Output.Tex.xy = float2((Input.VertexID << 1) & 2, Input.VertexID & 2);
    Output.Pos = float4(Output.Tex.xy * float2(2, -2) + float2(-1, 1), 0.5f, 1.f);
    
    //Store the adapted exposure in z coord of texture
    Output.Tex.z  = Exposure / max(LuminanceMaxAdaption, Luminance1x1Texture.SampleLevel(sampBilinear, float3(0.5f, 0.5f, 0.0f), 0).x);

    return Output;
}

//-------------------------------------------------------------

float4 Downscale_PS(in PSIn Input) : SV_Target
{
    float4 sample = 0.0f;

    return SourceTexture.Sample(sampBilinear, Input.TexCoord.xy);

}

//-------------------------------------------------------------

float4 GaussianBlur_PS(in PSIn Input) : SV_Target
{
    // Preshader
    float2 TexelSize = 1 / SourceSize;
    float  BlurSizeInv = 1 / float(GaussianBlurSamples);
    float2 SampleOffset = TexelSize * GaussianBlurDimension;
    float2 Offset = 0.5 * float(GaussianBlurSamples - 1) * SampleOffset;
    
    float2 BaseTexCoord = Input.TexCoord - Offset;
    float4 Sum = 0;
    
    for (int i = 0; i < GaussianBlurSamples; ++i) 
    {
        Sum += GaussianBlurWeights[i] * SourceTexture.SampleLevel(sampBilinear, BaseTexCoord + i * SampleOffset, 0);         
    }    
    
    return Sum;
}

//-------------------------------------------------------------

float4 LuminanceDownscaleLog_PS( in PSIn Input ) : SV_TARGET
{
    float luminance = 0; 
    float2 TexelSize = 1 / SourceSize;
    float Offset = 0.5 * TexelSize;
    float2 OffsetHor = float2(Offset,0);
    float2 OffsetVer = float2(0,Offset);
    
    float2 BaseTexCoordHor = Input.TexCoord - OffsetHor;
    float2 BaseTexCoordVer = Input.TexCoord - OffsetVer;


    for( int i = 0; i < 2; i++ )
    {
        luminance += log( dot( SourceTexture.Sample( sampBilinear, BaseTexCoordHor + i * OffsetHor).xyz, LUMINANCE_WEIGHTS ) + 0.0001f);
        luminance += log( dot( SourceTexture.Sample( sampBilinear, BaseTexCoordVer + i * OffsetVer).xyz, LUMINANCE_WEIGHTS ) + 0.0001f);    
        
    }

    luminance *= 0.25f;
    return float4(luminance, luminance, luminance, 1.0);
}

float4 LuminanceDownscale_PS( in PSIn Input ) : SV_TARGET
{
    float sample = 0.0f;
    float2 TexelSize = 1 / SourceSize;
    float Offset = 0.5 * TexelSize;
    float2 OffsetHor = float2(Offset,0);
    float2 OffsetVer = float2(0,Offset);
    
    float2 BaseTexCoordHor = Input.TexCoord - OffsetHor;
    float2 BaseTexCoordVer = Input.TexCoord - OffsetVer;
    
    for( int i = 0; i < 2; i++ )
    {
        sample += SourceTexture.Sample( sampBilinear, BaseTexCoordHor + i * OffsetHor ).x;
        sample += SourceTexture.Sample( sampBilinear, BaseTexCoordVer + i * OffsetVer ).x;
    }

    sample *= 0.25f;

    return float4(sample, sample, sample, 1.0);

}

float4 LuminanceDownscaleExp_PS( in PSIn Input ) : SV_TARGET
{
    float sample = 0.0f;

    float2 TexelSize = 1 / SourceSize;
    float Offset = 0.5 * TexelSize;
    float2 OffsetHor = float2(Offset,0);
    float2 OffsetVer = float2(0,Offset);
    
    float2 BaseTexCoordHor = Input.TexCoord - OffsetHor;
    float2 BaseTexCoordVer = Input.TexCoord - OffsetVer;

    for( int i = 0; i < 2; i++ )
    {
        sample += SourceTexture.Sample( sampBilinear, BaseTexCoordHor + i * OffsetHor ).x;
        sample += SourceTexture.Sample( sampBilinear, BaseTexCoordVer + i * OffsetVer ).x;
    }

    sample *= 0.25f;
	//sample = exp(sample);
    //return float4(sample, sample, sample, 1.0);
    return exp(sample);
    //return 1.0f;
}

//-------------------------------------------------------------

float4 LuminanceAdaptation_PS(in PSIn Input) : SV_TARGET
{
    float adaptedLum = SourceTexture.Sample(sampBilinear, float2(0.5f, 0.5f)).x;
    float currentLum = Luminance1x1Texture.Sample(sampBilinear, float2(0.5f, 0.5f)).x;

    return (adaptedLum + (currentLum - adaptedLum) * ( 1 - pow( 0.98f, ADAPTION_SPEED * TimeSinceLastFrame ) ));
}


//-------------------------------------------------------------


float4 BrightPassFilter_PS(in PSIn Input) : SV_TARGET
{
	float3 color = SourceTexture.Sample( sampBilinear, Input.TexCoord).rgb;
	
	color = max( 0.0f, color - BRIGHT_THRESHOLD );	

	return float4(color,1.0f);

}

//-------------------------------------------------------------

void vignette(inout float3 c, const float2 win_bias)
{
    // convert window coord to [-1, 1] range
    float2 wpos = 2*(win_bias - (float2)(0.5, 0.5)); 

    // calculate distance from origin
    float r = length(wpos.xy);
    r = 1.0 - smoothstep(0.8, 1.5, r);
    c = c * r;
}

float4 ToneMapping_PS(in PSInToneMapping Input) : SV_TARGET
{
    float3 c;
    

    // sum original and blurred image
     c = (1.0f - BlurWeight) * SourceTexture.Sample(sampBilinear, Input.Tex.xy).rgb + BlurWeight * BlurTexture.Sample(sampBilinear, Input.Tex.xy).rgb;

    float3 L = c * Input.Tex.z;

    // exposure
    c = L / (1 + L); 

    // vignette effect (brightness drop off with distance from center)
    //vignette(c, Input.Tex.xy); 

    return float4(c, 1.0);
    
    
    // When no SRGB texture do a gamma correction
    //return pow(float4(c, 1.0), 0.5);
}

//-------------------------------------------------------------

float4 Const_PS() : SV_Target
{
    return ConstValue;
}

//-------------------------------------------------------------

float4 BoxBlur_PS(PSIn Input) : SV_Target
{
    // Preshader
    float2 TexelSize = 1 / SourceSize;
    float  BlurSizeInv = 1 / float(BlurSamples);
    float2 SampleOffset = TexelSize * BlurDimension;
    float2 Offset = 0.5 * float(BlurSamples - 1) * SampleOffset;
    
    float2 BaseTexCoord = Input.TexCoord - Offset;
    
    float4 Sum = 0;
    for (int i = 0; i < BlurSamples; ++i) 
    {
        Sum += SourceTexture.SampleLevel(sampBilinear, BaseTexCoord + i * SampleOffset, 0);
    }
    
    return Sum * BlurSizeInv;
}

float4 BoxBlurArray_PS(PSIn Input) : SV_Target
{
    // Preshader
    float2 TexelSize = 1 / SourceSize;
    float  BlurSizeInv = 1 / float(BlurSamples);
    float2 SampleOffset = TexelSize * BlurDimension;
    float2 Offset = 0.5 * float(BlurSamples - 1) * SampleOffset;
    
    float2 BaseTexCoord = Input.TexCoord - Offset;
    
    float4 Sum = float4(0, 0, 0, 0);
    for (int i = 0; i < BlurSamples; ++i) {
        Sum += SourceTextureArray.SampleLevel(sampBilinear,
                                            float3(BaseTexCoord + i * SampleOffset, 0),
                                            0);
    }
    
    return Sum * BlurSizeInv;
}

//-------------------------------------------------------------

float4 BlendTexture_PS(PSIn input) : SV_Target
{
    // Preshader
	float4 Color = SourceTexture.Sample(sampBilinear, input.TexCoord);
	return float4 (Color.rgb, BlendFactor * Color.a);
}


//-------------------------------------------------------------
//-------------------------------------------------------------
technique10 BoxBlur
{
    pass p0
    {
        SetVertexShader(CompileShader(vs_4_0, PostProcess_VS()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, BoxBlur_PS()));
        
        SetRasterizerState(rsPostProcess);
        SetDepthStencilState(dsPostProcess, 0);
        SetBlendState(bsNormal, float4(0, 0, 0, 0), 0xFFFFFFFF);
    }
}

technique10 BoxBlurArray
{
    pass p0
    {
        SetVertexShader(CompileShader(vs_4_0, PostProcess_VS()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, BoxBlurArray_PS()));
        
        SetRasterizerState(rsPostProcess);
        SetDepthStencilState(dsPostProcess, 0);
        SetBlendState(bsNormal, float4(0, 0, 0, 0), 0xFFFFFFFF);
    }
}

technique10 Const
{
    pass p0
    {
        SetVertexShader(CompileShader(vs_4_0, PostProcess_VS()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, Const_PS()));
        
        SetRasterizerState(rsPostProcess);
        SetDepthStencilState(dsPostProcess, 0);
        SetBlendState(bsNormal, float4(0, 0, 0, 0), 0xFFFFFFFF);
    }
}

technique10 Downscale
{
    pass p0
    {
        SetVertexShader(CompileShader(vs_4_0, PostProcess_VS()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, Downscale_PS()));
        
        SetRasterizerState(rsPostProcess);
        SetDepthStencilState(dsPostProcess, 0);
        SetBlendState(bsNormal, float4(0, 0, 0, 0), 0xFFFFFFFF);
    }
}

technique10 GaussianBlur
{
    pass p0
    {
        SetVertexShader(CompileShader(vs_4_0, PostProcess_VS()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, GaussianBlur_PS()));
        //SetPixelShader(CompileShader(ps_4_0, BoxBlur_PS()));
        
        SetRasterizerState(rsPostProcess);
        SetDepthStencilState(dsPostProcess, 0);
        SetBlendState(bsNormal, float4(0, 0, 0, 0), 0xFFFFFFFF);
    }
}

technique10 LuminanceDownscaleLog
{
    pass p0
    {
        SetVertexShader(CompileShader(vs_4_0, PostProcess_VS()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, LuminanceDownscaleLog_PS()));
        
        SetRasterizerState(rsPostProcess);
        SetDepthStencilState(dsPostProcess, 0);
        SetBlendState(bsNormal, float4(0, 0, 0, 0), 0xFFFFFFFF);
    }
}

technique10 LuminanceDownscale
{
    pass p0
    {
        SetVertexShader(CompileShader(vs_4_0, PostProcess_VS()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, LuminanceDownscale_PS()));
        
        SetRasterizerState(rsPostProcess);
        SetDepthStencilState(dsPostProcess, 0);
        SetBlendState(bsNormal, float4(0, 0, 0, 0), 0xFFFFFFFF);
    }
}

technique10 LuminanceDownscaleExp
{
    pass p0
    {
        SetVertexShader(CompileShader(vs_4_0, PostProcess_VS()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, LuminanceDownscaleExp_PS()));
        
        SetRasterizerState(rsPostProcess);
        SetDepthStencilState(dsPostProcess, 0);
        SetBlendState(bsNormal, float4(0, 0, 0, 0), 0xFFFFFFFF);
    }
}

technique10 LuminanceAdaption
{
    pass p0
    {
        SetVertexShader(CompileShader(vs_4_0, PostProcess_VS()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, LuminanceAdaptation_PS()));
        
        SetRasterizerState(rsPostProcess);
        SetDepthStencilState(dsPostProcess, 0);
        SetBlendState(bsNormal, float4(0, 0, 0, 0), 0xFFFFFFFF);
    }
}

technique10 ToneMapping
{
    pass p0
    {
        SetVertexShader(CompileShader(vs_4_0, ToneMapping_VS()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, ToneMapping_PS()));
        
        SetRasterizerState(rsPostProcess);
        SetDepthStencilState(dsPostProcess, 0);
        SetBlendState(bsNormal, float4(0, 0, 0, 0), 0xFFFFFFFF);
    }
}

technique10 BrightPassFilter
{
    pass p0
    {
        SetVertexShader(CompileShader(vs_4_0, PostProcess_VS()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, BrightPassFilter_PS()));
        
        SetRasterizerState(rsPostProcess);
        SetDepthStencilState(dsPostProcess, 0);
        SetBlendState(bsNormal, float4(0, 0, 0, 0), 0xFFFFFFFF);
    }    
}

technique10 BlendTextureTechnique
{
    pass p0
    {
        SetVertexShader(CompileShader(vs_4_0, PostProcess_VS()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, BlendTexture_PS()));
        
        SetRasterizerState(rsPostProcess);
        SetDepthStencilState(dsPostProcess, 0);
        SetBlendState(SrcAlphaBlendingAdd, float4(0, 0, 0, 0), 0xFFFFFFFF);
        //SetBlendState(bsNormal, float4(0, 0, 0, 0), 0xFFFFFFFF);
    }    
}
