Texture2D Texture;
Texture1D ZHBlockerLookupTable;
Texture1D ZHLightLookupTable;
Texture1D ExpCoefficientsLookupTable;
TextureCube ZHRotationLookupTable1;
TextureCube ZHRotationLookupTable2;
TextureCube ZHRotationLookupTable3;
TextureCube ZHRotationLookupTable4;
int NumSpheres;
float4 Spheres[128];

#define EXP_MAG_MAX 2.974f
#define PI 3.141592f

SamplerState SamLinearClamp
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
    AddressW = Clamp;
};

struct SHCoeff
{
	float4 c1, c2, c3, c4;
};

SHCoeff RotateZH(float4 zh, float3 vec)
{
	SHCoeff rotated;
	
	// Lookup coeffs
	rotated.c1 = ZHRotationLookupTable1.Sample(SamLinearClamp, vec);
	rotated.c2 = ZHRotationLookupTable2.Sample(SamLinearClamp, vec);
	rotated.c3 = ZHRotationLookupTable3.Sample(SamLinearClamp, vec);
	rotated.c4 = ZHRotationLookupTable4.Sample(SamLinearClamp, vec);
	
	rotated.c1.x *= zh.x;
	rotated.c1.yzw *= zh.y;
	rotated.c2 *= zh.z;
	rotated.c3.x *= zh.z;
	rotated.c3.yzw *= zh.w;
	rotated.c4 *= zh.w;
	
	return rotated;
}

SHCoeff SHCalcDiffuseBRDF(float3 normal)
{
	// Speedup: don't read all rotation coeffs
	return RotateZH(float4(0, 2.04665f, 0, 0), normal);
}

float SHDot(SHCoeff sh1, SHCoeff sh2)
{
	return dot(sh1.c1, sh2.c1) + dot(sh1.c2, sh2.c2) + dot(sh1.c3, sh2.c3) + dot(sh1.c4, sh2.c4);
}

float SHMagnitude(SHCoeff sh)
{
	return dot(sh.c1, sh.c1) + dot(sh.c2, sh.c2) + dot(sh.c3, sh.c3) + dot(sh.c4, sh.c4);
}

SHCoeff SHExp(SHCoeff sh)
{
	float factor = exp(sh.c1.x / sqrt(4 * PI));
	sh.c1.x = 0;
	float2 ab = ExpCoefficientsLookupTable.Sample(SamLinearClamp, SHMagnitude(sh) / EXP_MAG_MAX) * factor;
	sh.c1 *= ab.y;
	sh.c2 *= ab.y;
	sh.c3 *= ab.y;
	sh.c4 *= ab.y;
	
	sh.c1.x = sqrt(4 * PI) * ab.x;
	
	return sh;
}

SHCoeff SHAdd(SHCoeff sh1, SHCoeff sh2)
{
	SHCoeff res;
	res.c1 = sh1.c1 + sh2.c1;
	res.c2 = sh1.c2 + sh2.c2;
	res.c3 = sh1.c3 + sh2.c3;
	res.c4 = sh1.c4 + sh2.c4;
	return res;
}

SHCoeff SHSquare(SHCoeff coeff) 
{
  float a, b; 
  SHCoeff res;

  res.c1.x = 0.2820948 * coeff.c1.x * coeff.c1.x;

  a = 0.5637281 * coeff.c3.z;
  res.c1.x += a * coeff.c3.z;
  res.c1.x -= 0.2818641 * coeff.c3.z * coeff.c3.z;
  res.c3.z = a * coeff.c1.x;

  a = 0.5638290 * coeff.c3.w;
  res.c1.x  += a * coeff.c3.w;
  res.c1.x -= 0.2819145 * coeff.c3.w * coeff.c3.w;
  res.c3.w = a * coeff.c1.x;

  a = 0.5646965 * coeff.c4.x;
  res.c1.x += a * coeff.c4.x;
  res.c1.x -= 0.2823482 * coeff.c4.x * coeff.c4.x;
  res.c4.x = a * coeff.c1.x;

  a = 0.5646505 * coeff.c4.y;
  res.c1.x += a * coeff.c4.y;
  res.c1.x -= 0.2823252 * coeff.c4.y * coeff.c4.y;
  res.c4.y = a * coeff.c1.x;

  a = 0.5645297 * coeff.c4.z;
  res.c1.x += a * coeff.c4.z;
  res.c1.x -= 0.2822649 * coeff.c4.z * coeff.c4.z;
  res.c4.z = a * coeff.c1.x;

  a = 0.5641159 * coeff.c1.x - 0.2522483 * coeff.c2.z - 0.4370127 * coeff.c3.x;
  res.c1.y = a * coeff.c1.y;
  b = coeff.c1.y * coeff.c1.y;
  res.c1.x += 0.2820580 * b;
  res.c2.z = -0.1261242 * b;
  res.c3.x = -0.2185064 * b;

  a = 0.4368935 * coeff.c1.w - 0.1169194 * coeff.c4.y - 0.4529205 * coeff.c4.w;
  res.c1.y  += a * coeff.c2.x;
  res.c2.x = a * coeff.c1.y;
  b = coeff.c1.y * coeff.c2.x;
  res.c1.w = 0.4368935 * b;
  res.c4.y -= 0.1169194 * b;
  res.c4.w = -0.4529205 * b;

  a = 0.4369993 * coeff.c1.z - 0.2862812 * coeff.c4.x - 0.3697703 * coeff.c4.z;
  res.c1.y  += a * coeff.c2.y;
  res.c2.y = a * coeff.c1.y;
  b = coeff.c1.y * coeff.c2.y;
  res.c1.z = 0.4369993 * b;
  res.c4.x  -= 0.2862812 * b;
  res.c4.z  -= 0.3697703 * b;

  a = 0.4516111 * coeff.c3.y;
  res.c1.y  += a * coeff.c3.x;
  res.c3.x  += a * coeff.c1.y;
  b = coeff.c1.y * coeff.c3.x;
  res.c3.y = 0.4516111 * b;

  a = 0.5645295 * coeff.c1.x + 0.5053624 * coeff.c2.z;
  res.c1.z  += a * coeff.c1.z;
  b = coeff.c1.z * coeff.c1.z;
  res.c1.x  += 0.2822647 * b;
  res.c2.z  += 0.2526812 * b;

  a = 0.5639233 * coeff.c1.x - 0.2519740 * coeff.c2.z + 0.4366057 * coeff.c3.x;
  res.c1.w  += a * coeff.c1.w;
  b = coeff.c1.w * coeff.c1.w;
  res.c1.x  += 0.2819617 * b;
  res.c2.z  -= 0.1259870 * b;
  res.c3.x  += 0.2183028 * b;

  a = 0.4370674 * coeff.c1.z - 0.2856638 * coeff.c4.x + 0.3693726 * coeff.c4.z;
  res.c1.w  += a * coeff.c2.w;
  res.c2.w = a * coeff.c1.w;
  b = coeff.c1.w * coeff.c2.w;
  res.c1.z  += 0.4370674 * b;
  res.c4.x  -= 0.2856638 * b;
  res.c4.z  += 0.3693726 * b;

  a = 0.4525065 * coeff.c4.w;
  res.c1.w  += a * coeff.c3.x;
  res.c3.x  += a * coeff.c1.w;
  b = coeff.c1.w * coeff.c3.x;
  res.c4.w  += 0.4525065 * b;

  a = 0.5640270 * coeff.c1.x - 0.3604869 * coeff.c2.z;
  res.c2.x  += a * coeff.c2.x;
  b = coeff.c2.x * coeff.c2.x;
  res.c1.x  += 0.2820135 * b;
  res.c2.z  -= 0.1802434 * b;

  a = 0.3119014 * coeff.c2.w;
  res.c2.x  += a * coeff.c2.y;
  res.c2.y  += a * coeff.c2.x;
  b = coeff.c2.x * coeff.c2.y;
  res.c2.w  += 0.3119014 * b;

  a = 0.4516148 * coeff.c1.w - 0.18772 * coeff.c4.y;
  res.c2.x  += a * coeff.c3.y;
  res.c3.y  += a * coeff.c2.x;
  b = coeff.c2.x * coeff.c3.y;
  res.c1.w  += 0.4516148 * b;
  res.c4.y  -= 0.18772 * b;

  a = 0.3690467 * coeff.c1.z - 0.3759367 * coeff.c4.x;
  res.c2.x  += a * coeff.c3.z;
  res.c3.z  += a * coeff.c2.x;
  b = coeff.c2.x * coeff.c3.z;
  res.c1.z  += 0.3690467 * b;
  res.c4.x  -= 0.3759367 * b;

  a = -0.1169194 * coeff.c1.w + 0.2911963 * coeff.c4.y + 0.1885629 * coeff.c4.w;
  res.c2.x  += a * coeff.c3.w;
  res.c3.w  += a * coeff.c2.x;
  b = coeff.c2.x * coeff.c3.w;
  res.c1.w  -= 0.1169194 * b;
  res.c4.y  += 0.2911963 * b;
  res.c4.w  += 0.1885629 * b;

  a = 0.5641636 * coeff.c1.x + 0.1799909 * coeff.c2.z - 0.3125129 * coeff.c3.x;
  res.c2.y  += a * coeff.c2.y;
  b = coeff.c2.y * coeff.c2.y;
  res.c1.x  += 0.2820818 * b;
  res.c2.z  += 0.0899955 * b;
  res.c3.x  -= 0.1562565 * b;

  a = 0.2975015 * coeff.c4.z;
  res.c2.y  += a * coeff.c3.y;
  res.c3.y  += a * coeff.c2.y;
  b = coeff.c2.y * coeff.c3.y;
  res.c4.z  += 0.2975015 * b;

  a = 0.3690467 * coeff.c1.w + 0.2299966 * coeff.c4.y - 0.2973875 * coeff.c4.w;
  res.c2.y  += a * coeff.c3.z;
  res.c3.z  += a * coeff.c2.y;
  b = coeff.c2.y * coeff.c3.z;
  res.c1.w  += 0.3690467 * b;
  res.c4.y  += 0.2299966 * b;
  res.c4.w  -= 0.2973875 * b;

  a = 0.4669293 * coeff.c1.z + 0.1188663 * coeff.c4.x - 0.2304082 * coeff.c4.z;
  res.c2.y  += a * coeff.c3.w;
  res.c3.w  += a * coeff.c2.y;
  b = coeff.c2.y * coeff.c3.w;
  res.c1.z  += 0.4669293 * b;
  res.c4.x  += 0.1188663 * b;
  res.c4.z  -= 0.2304082 * b;

  a = 0.5645874 * coeff.c1.x;
  res.c2.z  += a * coeff.c2.z;
  b = coeff.c2.z * coeff.c2.z;
  res.c1.x  += 0.2822937 * b;
  res.c2.z  += 0.1808038 * b;

  a = 0.4043228 * coeff.c1.y + 0.2522087 * coeff.c3.w;
  res.c2.z  += a * coeff.c3.w;
  res.c2.z -= 0.1261043 * coeff.c3.w * coeff.c3.w;
  res.c3.w  += a * coeff.c2.z;
  b = coeff.c2.z * coeff.c3.w;
  res.c1.y  += 0.4043228 * b;

  a = 0.4964184 * coeff.c1.z + 0.3372459 * coeff.c4.x;
  res.c2.z  += a * coeff.c4.x;
  res.c2.z -= 0.1686229 * coeff.c4.x * coeff.c4.x;
  res.c4.x  += a * coeff.c2.z;
  b = coeff.c2.z * coeff.c4.x;
  res.c1.z  += 0.4964184 * b;

  a = 0.4047946 * coeff.c1.w + 0.2538309 * coeff.c4.y;
  res.c2.z  += a * coeff.c4.y;
  res.c2.z -= 0.1269154 * coeff.c4.y * coeff.c4.y;
  res.c4.y  += a * coeff.c2.z;
  b = coeff.c2.z * coeff.c4.y;
  res.c1.w  += 0.4047946 * b;

  a = 0.5642515 * coeff.c1.x + 0.1807696 * coeff.c2.z + 0.3121768 * coeff.c3.x;
  res.c2.w  += a * coeff.c2.w;
  b = coeff.c2.w * coeff.c2.w;
  res.c1.x  += 0.2821258 * b;
  res.c2.z  += 0.0903848 * b;
  res.c3.x  += 0.1560884 * b;

  a = 0.3690468 * coeff.c1.y + 0.2969087 * coeff.c3.y + 0.2299966 * coeff.c3.w;
  res.c2.w  += a * coeff.c3.z;
  res.c3.z  += a * coeff.c2.w;
  b = coeff.c2.w * coeff.c3.z;
  res.c1.y  += 0.3690467 * b;
  res.c3.y  += 0.2969087 * b;
  res.c3.w  += 0.2299966 * b;

  a = 0.4678128 * coeff.c1.z + 0.1199120 * coeff.c4.x + 0.2306436 * coeff.c4.z;
  res.c2.w  += a * coeff.c4.y;
  res.c4.y  += a * coeff.c2.w;
  b = coeff.c2.w * coeff.c4.y;
  res.c1.z  += 0.4678128 * b;
  res.c4.x  += 0.119912 * b;
  res.c4.z  += 0.2306436 * b;

  a = 0.2975354 * coeff.c4.w;
  res.c2.w  += a * coeff.c4.z;
  res.c4.z  += a * coeff.c2.w;
  b = coeff.c2.w * coeff.c4.z;
  res.c4.w  += 0.2975354 * b;

  a = 0.5639183 * coeff.c1.x - 0.3599812 * coeff.c2.z;
  res.c3.x  += a * coeff.c3.x;
  b = coeff.c3.x * coeff.c3.x;
  res.c1.x  += 0.2819591 * b;
  res.c2.z  -= 0.1799906 * b;

  a = 0.1164589 * coeff.c1.y - 0.1872479 * coeff.c3.y - 0.2910910 * coeff.c3.w;
  res.c3.x  += a * coeff.c3.w;
  res.c3.x += 0.1455455 * coeff.c3.w * coeff.c3.w;
  res.c3.w  += a * coeff.c3.x;
  b = coeff.c3.x * coeff.c3.w;
  res.c1.y  += 0.1164589 * b;
  res.c3.y  -= 0.1872479 * b;

  a = -0.1163925 * coeff.c1.w + 0.291215 * coeff.c4.y - 0.1880587 * coeff.c4.w;
  res.c3.x  += a * coeff.c4.y;
  res.c3.x -= 0.1456075 * coeff.c4.y * coeff.c4.y;
  res.c4.y  += a * coeff.c3.x;
  b = coeff.c3.x * coeff.c4.y;
  res.c1.w  -= 0.1163925 * b;
  res.c4.w  -= 0.1880587 * b;

  a = 0.3695715 * coeff.c1.z - 0.3763061 * coeff.c4.x;
  res.c3.x  += a * coeff.c4.z;
  res.c4.z  += a * coeff.c3.x;
  b = coeff.c3.x * coeff.c4.z;
  res.c1.z  += 0.3695715 * b;
  res.c4.x  -= 0.3763061 * b;

  a = 0.5632603 * coeff.c1.x - 0.4195884 * coeff.c2.z;
  res.c3.y  += a * coeff.c3.y;
  b = coeff.c3.y * coeff.c3.y;
  res.c1.x  += 0.2816301 * b;
  res.c2.z  -= 0.2097942 * b;

  a = 0.564633 * coeff.c1.x - 0.4209418 * coeff.c2.z;
  res.c4.w  += a * coeff.c4.w;
  b = coeff.c4.w * coeff.c4.w;
  res.c1.x  += 0.2823165 * b;
  res.c2.z  -= 0.2104709 * b;

  return res;
}

float CalcSHLighting(float3 lightPos, float lightRadius, float lightStrength, float3 vertexInWorld, float3 vertexNormalWorld)
{
	float3 dir = lightPos - vertexInWorld;
	float lenToLight = length(dir);
	
	SHCoeff light = RotateZH(ZHLightLookupTable.Sample(SamLinearClamp, (lightRadius / lenToLight) / (PI / 2)), dir);
	light.c1 *= lightStrength;
	light.c2 *= lightStrength;
	light.c3 *= lightStrength;
	light.c4 *= lightStrength;

	SHCoeff finalSH;
	finalSH.c1 = 0;
	finalSH.c2 = 0;
	finalSH.c3 = 0;
	finalSH.c4 = 0;
	
	for(int i=0 ; i < NumSpheres ; i++)
	{
		float3 center = Spheres[i].xyz;
		float radius = Spheres[i].w;

		dir = center - vertexInWorld;
		
		float3 lightPlane = center - lightPos;
		
		if(dot(lightPlane, vertexNormalWorld) < 0)
		{
			float planeDist = dot(dir, vertexNormalWorld);
			
			if(abs(planeDist) < radius)
			{
				if (planeDist >= 0.0f)
				{
					radius = planeDist;
					center = (vertexNormalWorld * ((radius - planeDist) / 2.0f)) + center;
				}
				else
				{
					float d2sq = dot(dir, dir);

					float3 q1 = (vertexNormalWorld * -planeDist) + center;

					float d = sqrt(radius * radius - planeDist * planeDist);
					float scale = min((length(vertexInWorld - q1) - d) / d, 1.0f);

					radius = (radius + planeDist) * 0.5 * scale;
					center = q1 + vertexNormalWorld * radius;
				}
				dir = center - vertexInWorld;	
			}
			float len = length(dir);
				
			SHCoeff res = RotateZH(ZHBlockerLookupTable.Sample(SamLinearClamp, (radius / len) / (PI / 2) + 0.042f), dir);
			finalSH = SHAdd(finalSH, res);
		}
	}
	
	int nsq = 2;
	float val= pow(0.5, nsq);
	finalSH.c1 *= val;
	finalSH.c2 *= val;
	finalSH.c3 *= val;
	finalSH.c4 *= val;
	
	finalSH = SHExp(finalSH);
	
	for(int i=0 ; i < nsq ; i++)
		finalSH = SHSquare(finalSH);
		
	float shade = min(max(SHDot(finalSH, light), 0.01f), 1);
	return shade;
}