matrix World;
matrix View;
matrix Projection;
matrix WorldViewProjection;
float3 LightDir[3];
float3 SHLight;
float3 LightColor[3];
float3 AmbientColor;
float3 DiffuseColor;
float3 SpecularColor;
float3 EmissiveColor;
float3 CameraPos;
float Shininess;
int NumLights;
float Opacity;
matrix BoneMatrices[64];

float3 ClipPlaneNormal;
float  ClipPlaneDistance;

bool bParallaxMapping;

Texture2D NormalMap;
Texture2D LightMap;

static const float PARALLAXHEIGHT = 0.06f;
static const float PARALLAXOFFSET = -0.01f;

BlendState NoAlpha
{
    BlendEnable[0] = FALSE;
};

BlendState SrcAlphaBlendingAdd
{
    BlendEnable[0] = TRUE;
    SrcBlend = SRC_ALPHA;
    DestBlend = INV_SRC_ALPHA;
    BlendOp = ADD;
};

#include "SHLighting.fx"
#include "ShadowMap.fx"

RasterizerState DisableCulling
{
    CullMode = NONE;
};

//--------------------------------------------------------------------------------------
struct VS_INPUT
{
    float4 Pos : POSITION;
    float3 Norm : NORMAL;
	float2 Texcoord : TEXCOORD0;
    float4 Tangent	: TANGENT0; // pre-normalized
    float3 Binormal	: BINORMAL0; // pre-normalized
};

struct VS_INPUT_LIGHTMAP
{
    float4 Pos : POSITION;
    float3 Norm : NORMAL;
	float2 Texcoord : TEXCOORD0;
	float2 Texcoord2 : TEXCOORD1;
    float4 Tangent	: TANGENT0; // pre-normalized
    float3 Binormal	: BINORMAL0; // pre-normalized
};

struct VS_INPUT_SKIN
{
    float4 Pos : POSITION;
    float3 Norm : NORMAL;
	float2 Texcoord : TEXCOORD0;
    float3 Tangent	: TANGENT0; // pre-normalized
    float3 Binormal	: BINORMAL0; // pre-normalized
	uint4  Indices : BLENDINDICES;
    float4 Weights : BLENDWEIGHTS;
};

struct VS_INPUT_NOTEXTURE
{
    float4 Pos : POSITION;
    float3 Norm : NORMAL;
    float3 Tangent	: TANGENT0; // pre-normalized
    float3 Binormal	: BINORMAL0; // pre-normalized    
};

struct VS_INPUT_NOTEXTURE_SKIN
{
    float4 Pos : POSITION;
    float3 Norm : NORMAL;
    float3 Tangent	: TANGENT0; // pre-normalized
    float3 Binormal	: BINORMAL0; // pre-normalized    
    uint4  Indices : BLENDINDICES;
    float4 Weights : BLENDWEIGHTS;
};
  
struct PS_INPUT
{
    float4 Pos : SV_POSITION;
	float2 Texcoord   : TEXCOORD0;
	float3 PosInWorld : TEXCOORD1;
    float3 ViewTS    : TEXCOORD2;   // view vector in tangent space, denormalized
    float3 LightTS1  : TEXCOORD3;   // light vector in tangent space for first light, denormalized
    float3 LightTS2  : TEXCOORD4;   // light vector in tangent space for second light, denormalized
    float3 LightTS3  : TEXCOORD5;   // light vector in tangent space for third light, denormalized
    float3 NormWS    : TEXCOORD6;
	float  Distance : SV_ClipDistance0;
	float  SliceDepth : SliceDepth; 
};

struct PS_INPUT_LIGHTMAP
{
    float4 Pos : SV_POSITION;
	float2 Texcoord   : TEXCOORD0;
	float2 Texcoord2   : TEXCOORD1;
	float3 PosInWorld : TEXCOORD2;
    float3 ViewTS    : TEXCOORD3;   // view vector in tangent space, denormalized
    float3 LightTS1  : TEXCOORD4;   // light vector in tangent space for first light, denormalized
    float3 LightTS2  : TEXCOORD5;   // light vector in tangent space for second light, denormalized
    float3 LightTS3  : TEXCOORD6;   // light vector in tangent space for third light, denormalized
    float3 NormWS    : TEXCOORD7;
	float  Distance : SV_ClipDistance0;
	float  SliceDepth : SliceDepth; 
};


struct PS_INPUT_NOTEXTURE
{
    float4 Pos : SV_POSITION;
    float3 PosInWorld : TEXCOORD0;    
    float3 NormWS : TEXCOORD1;
    float  Distance : SV_ClipDistance0;
    float  SliceDepth : SliceDepth; 
};

SamplerState SamLinear
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Wrap;
    AddressV = Wrap;
    AddressW = Wrap;
};

SamplerState SamAnisotropicWrap
{
    AddressU = Wrap;
    AddressV = Wrap;
    AddressW = Wrap;
    Filter = ANISOTROPIC;
    MaxAnisotropy = 8;
};

//--------------------------------------------------------------------------------------

PS_INPUT VS(VS_INPUT input)
{
    PS_INPUT output;
	
	output.PosInWorld = mul(input.Pos, World);
    output.Pos = mul(input.Pos, WorldViewProjection);
	output.Texcoord = input.Texcoord;
    
    output.NormWS = normalize (mul(input.Norm, (float3x3)World));
    float3 TangentWS = normalize (mul(input.Tangent, (float3x3)World));
    float3 BinormalWS = normalize (mul(input.Binormal, (float3x3)World));
    
    BinormalWS = cross(output.NormWS, TangentWS);
    
    float3 ViewWS = CameraPos - output.PosInWorld;
      
    float3x3 WorldToTangent = float3x3( TangentWS, BinormalWS, output.NormWS );
    
    output.LightTS1 = mul( WorldToTangent, LightDir[0] );
    output.LightTS2 = mul( WorldToTangent, LightDir[1] );
    output.LightTS3 = mul( WorldToTangent, LightDir[2] );
    output.ViewTS  = mul( WorldToTangent, ViewWS );
    
    output.Distance = ClipPlaneNormal.x * output.PosInWorld.x + ClipPlaneNormal.y *
    output.PosInWorld.y + ClipPlaneNormal.z * output.PosInWorld.z + ClipPlaneDistance;
      
	output.SliceDepth = output.Pos.z;      
      
    return output;
}

PS_INPUT_LIGHTMAP VS_LightMap(VS_INPUT_LIGHTMAP input)
{
    PS_INPUT_LIGHTMAP output;
	
	output.PosInWorld = mul(input.Pos, World);
    output.Pos = mul(input.Pos, WorldViewProjection);
	output.Texcoord = input.Texcoord;
	output.Texcoord2 = input.Texcoord2;
    
    output.NormWS = normalize (mul(input.Norm, (float3x3)World));
    float3 TangentWS = normalize (mul(input.Tangent, (float3x3)World));
    float3 BinormalWS = normalize (mul(input.Binormal, (float3x3)World));
    
    float3 ViewWS = CameraPos - output.PosInWorld;
      
    float3x3 WorldToTangent = float3x3( TangentWS, BinormalWS, output.NormWS );
    output.LightTS1 = mul( WorldToTangent, LightDir[0] );
    output.LightTS2 = mul( WorldToTangent, LightDir[1] );
    output.LightTS3 = mul( WorldToTangent, LightDir[2] );
    output.ViewTS  = mul( WorldToTangent, ViewWS );
    
    output.Distance = ClipPlaneNormal.x * output.PosInWorld.x + ClipPlaneNormal.y *
    output.PosInWorld.y + ClipPlaneNormal.z * output.PosInWorld.z + ClipPlaneDistance;
      
	output.SliceDepth = output.Pos.z;      
      
    return output;
}


PS_INPUT VS_Skin(VS_INPUT_SKIN input)
{
    PS_INPUT output;
    
    float4 skinnedPos = 0;
    float3 skinnedNormal = 0;
    for(int bone = 0; bone < 4; bone++)
    {
		skinnedPos += mul(input.Pos, BoneMatrices[input.Indices[bone]]) * input.Weights[bone];
		skinnedNormal += mul(input.Norm, (float3x3)BoneMatrices[input.Indices[bone]]) * input.Weights[bone];
	}

    output.PosInWorld = mul(skinnedPos, World);
    output.Pos = mul(skinnedPos, WorldViewProjection);    
	output.Texcoord = input.Texcoord;
    
    output.NormWS = normalize (mul(skinnedNormal, (float3x3)World));
    float3 TangentWS = normalize (mul(input.Tangent, (float3x3)World));
    float3 BinormalWS = normalize (mul(input.Binormal, (float3x3)World));
    
    float3 ViewWS = CameraPos - output.PosInWorld;
      
    float3x3 WorldToTangent = float3x3( TangentWS, BinormalWS, output.NormWS );
    output.LightTS1 = mul( WorldToTangent, LightDir[0] );
    output.LightTS2 = mul( WorldToTangent, LightDir[1] );
    output.LightTS3 = mul( WorldToTangent, LightDir[2] );
    output.ViewTS  = mul( WorldToTangent, ViewWS );
    
    output.Distance = ClipPlaneNormal.x * output.PosInWorld.x + ClipPlaneNormal.y *
    output.PosInWorld.y + ClipPlaneNormal.z * output.PosInWorld.z + ClipPlaneDistance;
      
	output.SliceDepth = output.Pos.z;      
      
    return output;
}

PS_INPUT_NOTEXTURE VS_NoTexture(VS_INPUT_NOTEXTURE input)
{
    PS_INPUT_NOTEXTURE output;
    output.PosInWorld = mul(input.Pos, World);
    output.Pos = mul(input.Pos, WorldViewProjection);

    output.NormWS = normalize (mul(input.Norm, (float3x3)World));
    
    output.Distance = ClipPlaneNormal.x * output.PosInWorld.x + ClipPlaneNormal.y *
    output.PosInWorld.y + ClipPlaneNormal.z * output.PosInWorld.z + ClipPlaneDistance;

	output.SliceDepth = output.Pos.z;

    return output;
}

PS_INPUT_NOTEXTURE VS_NoTexture_Skin(VS_INPUT_NOTEXTURE_SKIN input)
{
    PS_INPUT_NOTEXTURE output;
    float4 skinnedPos = 0;
    float3 skinnedNormal = 0;
    for(int bone = 0; bone < 4; bone++)
    {
		skinnedPos += mul(input.Pos, BoneMatrices[input.Indices[bone]]) * input.Weights[bone];
		skinnedNormal += mul(input.Norm, (float3x3)BoneMatrices[input.Indices[bone]]) * input.Weights[bone];
	}
    
    output.PosInWorld = mul(skinnedPos, World);
    output.Pos = mul(skinnedPos, WorldViewProjection);
    output.NormWS = normalize (mul(skinnedNormal, (float3x3)World));
    
    output.Distance = ClipPlaneNormal.x * output.PosInWorld.x + ClipPlaneNormal.y *
    output.PosInWorld.y + ClipPlaneNormal.z * output.PosInWorld.z + ClipPlaneDistance;

	output.SliceDepth = output.Pos.z;
    
    return output;
}

//--------------------------------------------------------------------------------------
// Color and normal calculation---------------------------------------------------------
//--------------------------------------------------------------------------------------

/**
 * Samples and Calculates normal value and modified texCoords
 * @input texCoord Texture Coordinates of NormalMap
 * @input viewTS View Vector in Tangent Space (Normalized!)
 * @input out texCoordMod modified texture coordinates
 * @return Normal Vector sampled from Normal Map (in Tangent Space)
 **/
float3 GetParallaxMappingValues(in float2 texCoord, in float3 ViewTS, out float2 texCoordMod)
{
   // Sample the height map at the current texture coordinate:
   //float4 parallaxValue = NormalMap.SampleLevel(SamLinear,texCoord, 0);
   float4 parallaxValue = NormalMap.Sample(SamAnisotropicWrap, texCoord);
   float fCurrentHeight =  parallaxValue.a;
   
   // Scale and bias this height map value:
   float fHeight = fCurrentHeight * PARALLAXHEIGHT + PARALLAXOFFSET;
   
   // Perform offset limiting if desired:
   //fHeight /= ViewTS.z;
   
   // Compute the offset vector for approximating parallax:
   texCoordMod = texCoord; //+ ViewTS * fHeight;
   
   // Return normal from normal-map
   return normalize(parallaxValue.xyz * 2 - 1);
}

float3 CalculateDiffuseColor(float3 norm, float3 ShadowContrib, float3 LightMapContrib)
{
    float3 finalColor = 0;

    //do NdotL lighting for 2 lights
    for(int light = 0 ; light < NumLights ; light++)
	{
        float3 ambient = float3(AmbientColor.rgb * LightColor[light].rgb);
		float nDotL  = clamp(dot(-LightDir[light], norm),0.0,1.0);
		float3 diffuse = float3(DiffuseColor.rgb * LightColor[light].rgb) * nDotL;
		//finalColor += saturate(ambient + ShadowContrib * diffuse);
		finalColor += LightMapContrib * (ambient + ShadowContrib * diffuse);
	}
	
	finalColor += EmissiveColor;
	return (finalColor);
}

float3 CalculateSpecularColor(float3 norm, float3 posInWorld, float3 ShadowContrib, float3 LightMapContrib)
{
    float3 finalColor = 0;
	
    //do NdotL lighting for 2 lights
    for(int light = 0 ; light < NumLights ; light++)
	{
        float3 ambient = float3(AmbientColor.rgb * LightColor[light].rgb);
		float nDotL  = clamp(dot(-LightDir[light], norm),0.0,1.0);
		float3 diffuse = float3(DiffuseColor.rgb * LightColor[light].rgb) * nDotL;
								
		//float3 halfAngle = normalize(normalize(CameraPos) - LightDir[light]);
		float3 view = CameraPos - posInWorld;
		float3 halfAngle = normalize( 2 * dot( view, norm ) * norm - view );
		float RdotL = saturate( dot( halfAngle, -LightDir[light] ));
		float3 specular = float3(SpecularColor.rgb * LightColor[light].rgb) * pow(RdotL, Shininess); 
		
		//finalColor += saturate(ambient + ShadowContrib * (diffuse + specular));
		finalColor += LightMapContrib * (ambient + ShadowContrib * (diffuse + specular));
	}
	finalColor += EmissiveColor;
	return (finalColor);
}

/**
 * Note NormalTS and ViewTS needs to be normalized
 * LightTS can be unnormalized
 **/
float3 CalculateDiffuseColorTS(float3 NormalTS, float3 ViewTS, float3 LightTS[3], float3 ShadowContrib, float3 LightMapContrib)
{
    float3 finalColor = 0;
	  	
    //do NdotL lighting for 2 lights
    for(int light = 0 ; light < NumLights ; light++)
	{
		float3 LightDirTS = normalize(-LightTS[light]);
		float3 LightAdj	= float3( LightDirTS.x, -LightDirTS.y, LightDirTS.z );
        float3 ambient = float3(AmbientColor.rgb * LightColor[light].rgb);
		float nDotL  = saturate(dot(LightAdj, NormalTS));
		float3 diffuse = float3(DiffuseColor.rgb * LightColor[light].rgb) * nDotL;
		finalColor += LightMapContrib * (ambient + ShadowContrib * diffuse);
	}
	
	finalColor += EmissiveColor;
	return (finalColor);
}

float3 CalculateSpecularColorTS(float3 NormalTS, float3 ViewTS, float3 LightTS[3], float3 ShadowContrib, float3 LightMapContrib)
{
    float3 finalColor = 0;

    //do NdotL lighting for 2 lights
    for(int light = 0 ; light < NumLights ; light++)
	{
		float3 LightDirTS = normalize(-LightTS[light]);
		float3 LightAdj	= float3( LightDirTS.x, -LightDirTS.y, LightDirTS.z );
        float3 ambient = float3(AmbientColor.rgb * LightColor[light].rgb);
		float nDotL  = saturate(dot(NormalTS, LightAdj));
		float3 diffuse = float3(DiffuseColor.rgb * LightColor[light].rgb) * nDotL;
								
		float3 halfAngle = normalize( 2 * dot( ViewTS, NormalTS ) * NormalTS - ViewTS );
		float RdotL = saturate( dot( halfAngle, LightAdj ));
		float3 specular = float3(SpecularColor.rgb * LightColor[light].rgb) * pow(RdotL, Shininess); 

		finalColor += LightMapContrib*(ambient + ShadowContrib * (diffuse + specular));
	}
	
	finalColor += EmissiveColor;
	
	return (finalColor);
}


//--------------------------------------------------------------------------------------
// PS Diffuse--------------------------------------------------------------------------
//--------------------------------------------------------------------------------------

float4 PSDiffuse(PS_INPUT input) : SV_Target
{
	if (bParallaxMapping)
	{	
		float2 newTexCoord;
		float3 ViewTS = normalize(input.ViewTS);
		float3 LightTS[3] = {input.LightTS1, input.LightTS2, input.LightTS3 };
		float3 NormTS = GetParallaxMappingValues(input.Texcoord, ViewTS, newTexCoord);
	
		float4 texCol = Texture.Sample(SamAnisotropicWrap, newTexCoord);
		return float4(texCol.rgb * CalculateDiffuseColorTS(NormTS, ViewTS, LightTS, float3(1.0,1.0,1.0), float3(1.0,1.0,1.0)).xyz, texCol.a * Opacity);
	}
	else
	{
		float4 texCol = Texture.Sample(SamAnisotropicWrap, input.Texcoord);
		return float4(texCol.rgb * CalculateDiffuseColor(normalize(input.NormWS), float3(1.0,1.0,1.0), float3(1.0,1.0,1.0)).xyz, texCol.a * Opacity);
	}	
}

float4 PSDiffuseLightMap(PS_INPUT_LIGHTMAP input) : SV_Target
{

	float3 lm = LightMap.Sample(SamAnisotropicWrap, input.Texcoord2).rgb;
	
	if (bParallaxMapping)
	{	
		float2 newTexCoord;
		float3 ViewTS = normalize(input.ViewTS);
		float3 LightTS[3] = {input.LightTS1, input.LightTS2, input.LightTS3 };
		float3 NormTS = GetParallaxMappingValues(input.Texcoord, ViewTS, newTexCoord);
	
		float4 texCol = Texture.Sample(SamAnisotropicWrap, newTexCoord);
		return float4(texCol.rgb * CalculateDiffuseColorTS(NormTS, ViewTS, LightTS, float3(1.0,1.0,1.0), lm).xyz, texCol.a * Opacity);
	}
	else
	{
		float4 texCol = Texture.Sample(SamAnisotropicWrap, input.Texcoord);
		return float4(texCol.rgb * CalculateDiffuseColor(normalize(input.NormWS), float3(1.0,1.0,1.0), lm).xyz, texCol.a * Opacity);
	}	
}

float4 PSDiffuseShadowMap(PS_INPUT input) : SV_Target
{
    // Calculate LightTexCoords by using Light ViewProjectionMatrix
    // and dividing by w
    int Split = GetSplit(input.SliceDepth);
    float4 PosLight = mul(float4(input.PosInWorld, 1), SplitMatrices[Split]);
    float2 LightTexCoord = (PosLight.xy / PosLight.w) * float2(0.5, -0.5) + 0.5;
    	
	float3 ShadowContrib = CalcShadowContribution(input.PosInWorld, LightTexCoord, Split);    

	if (bParallaxMapping)
	{
		float2 newTexCoord;
		float3 ViewTS = normalize(input.ViewTS);
		float3 LightTS[3] = {input.LightTS1, input.LightTS2, input.LightTS3 };
		float3 NormTS = GetParallaxMappingValues(input.Texcoord, ViewTS, newTexCoord);
		float4 texCol = Texture.Sample(SamAnisotropicWrap, newTexCoord);
	
		return float4(texCol.rgb * CalculateDiffuseColorTS(NormTS, ViewTS, LightTS, ShadowContrib, float3(1.0,1.0,1.0)).xyz, texCol.a * Opacity);		
	}
	else
	{
		float4 texCol = Texture.Sample(SamAnisotropicWrap, input.Texcoord);
		return float4(texCol.rgb * CalculateDiffuseColor(normalize(input.NormWS), ShadowContrib, float3(1.0,1.0,1.0)).xyz, texCol.a * Opacity);
	}
}

float4 PSSHDiffuse(PS_INPUT input) : SV_Target
{
	float4 texCol = Texture.Sample(SamAnisotropicWrap, input.Texcoord); 
	
	float shade = CalcSHLighting(SHLight, 12, 400, input.PosInWorld, input.NormWS);
	
	shade += 0.03f;
	  
	return float4(AmbientColor.rgb * texCol.rgb * min(1.0f, shade), texCol.a * Opacity);
}

float4 PSSHDiffuse_NoTexture(PS_INPUT_NOTEXTURE input) : SV_Target
{
	float shade = CalcSHLighting(SHLight, 12, 400, input.PosInWorld, input.NormWS);
	
	return float4(DiffuseColor * shade, Opacity);
}

float4 PSDiffuse_NoTexture(PS_INPUT_NOTEXTURE input) : SV_Target
{
	float3 norm = normalize(input.NormWS);
	return float4(CalculateDiffuseColor(norm, float3(1.0,1.0,1.0), float3(1.0,1.0,1.0)).rgb , Opacity);
}

float4 PSDiffuse_NoTextureShadowMap(PS_INPUT_NOTEXTURE input) : SV_Target
{
	float3 norm = normalize(input.NormWS);
	
    // Calculate LightTexCoords by using Light ViewProjectionMatrix
    // and dividing by w
    int Split = GetSplit(input.SliceDepth);
    float4 PosLight = mul(float4(input.PosInWorld, 1), SplitMatrices[Split]);
    float2 LightTexCoord = (PosLight.xy / PosLight.w) * float2(0.5, -0.5) + 0.5;
    	
	float3 ShadowContrib = CalcShadowContribution(input.PosInWorld, LightTexCoord, Split);	
	return float4(CalculateDiffuseColor(norm, ShadowContrib, float3(1.0,1.0,1.0)).rgb , Opacity);
}

//--------------------------------------------------------------------------------------
// PS Specular--------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
float4 PSSpecular(PS_INPUT input) : SV_Target
{
	if (bParallaxMapping)
	{
		float2 newTexCoord;
		float3 ViewTS = normalize(input.ViewTS);
		float3 LightTS[3] = {input.LightTS1, input.LightTS2, input.LightTS3 };
		float3 NormTS = GetParallaxMappingValues(input.Texcoord, ViewTS, newTexCoord);
    
		float4 texCol = Texture.Sample(SamAnisotropicWrap, newTexCoord);

		return float4(texCol.rgb * CalculateSpecularColorTS(NormTS, ViewTS, LightTS, float3(1.0,1.0,1.0), float3(1.0,1.0,1.0)).rgb, texCol.a * Opacity);
	}
	else
	{
		float4 texCol = Texture.Sample(SamAnisotropicWrap, input.Texcoord);
		return float4(texCol.rgb * CalculateSpecularColor(normalize(input.NormWS), input.PosInWorld, float3(1.0,1.0,1.0), float3(1.0,1.0,1.0)).rgb, texCol.a * Opacity);
	}		
}

float4 PSSpecularLightMap(PS_INPUT_LIGHTMAP input) : SV_Target
{

	float3 lightMap = LightMap.Sample(SamAnisotropicWrap, input.Texcoord2).rgb;
	
	if (bParallaxMapping)
	{
		float2 newTexCoord;
		float3 ViewTS = normalize(input.ViewTS);
		float3 LightTS[3] = {input.LightTS1, input.LightTS2, input.LightTS3 };
		float3 NormTS = GetParallaxMappingValues(input.Texcoord, ViewTS, newTexCoord);
    
		float4 texCol = Texture.Sample(SamAnisotropicWrap, newTexCoord);

		return float4(texCol.rgb * CalculateSpecularColorTS(NormTS, ViewTS, LightTS, float3(1.0,1.0,1.0), lightMap).rgb, texCol.a * Opacity);
	}
	else
	{
		float4 texCol = Texture.Sample(SamAnisotropicWrap, input.Texcoord);
		return float4(texCol.rgb * CalculateSpecularColor(normalize(input.NormWS), input.PosInWorld, float3(1.0,1.0,1.0), lightMap).rgb, texCol.a * Opacity);
	}		
}

float4 PSSpecularShadowMap(PS_INPUT input) : SV_Target
{
    // Calculate LightTexCoords by using Light ViewProjectionMatrix
    // and dividing by w
    int Split = GetSplit(input.SliceDepth);
    float4 PosLight = mul(float4(input.PosInWorld, 1), SplitMatrices[Split]);
    float2 LightTexCoord = (PosLight.xy / PosLight.w) * float2(0.5, -0.5) + 0.5;
	float3 ShadowContrib = CalcShadowContribution(input.PosInWorld, LightTexCoord, Split);

	if (bParallaxMapping)
	{
		float2 newTexCoord;
		float3 ViewTS = normalize(input.ViewTS);
		float3 LightTS[3] = {input.LightTS1, input.LightTS2, input.LightTS3 };
		float3 NormTS = GetParallaxMappingValues(input.Texcoord, ViewTS, newTexCoord);
		float4 texCol = Texture.Sample(SamAnisotropicWrap, newTexCoord);
    	
		return float4(texCol.rgb * CalculateSpecularColorTS(NormTS, ViewTS, LightTS, ShadowContrib, float3(1.0,1.0,1.0)).rgb, texCol.a * Opacity);		
	}
	else
	{
		float4 texCol = Texture.Sample(SamAnisotropicWrap, input.Texcoord);
    	return float4(texCol.rgb * CalculateSpecularColor(normalize(input.NormWS), input.PosInWorld, ShadowContrib, float3(1.0,1.0,1.0)).rgb, texCol.a * Opacity);
    }

   //return float4(ShadowContrib.xyz, Opacity);    
}

float4 PSSpecular_NoTexture(PS_INPUT_NOTEXTURE input) : SV_Target
{
	float3 norm = normalize(input.NormWS);
    return float4(CalculateSpecularColor(norm, input.PosInWorld, float3(1.0,1.0,1.0), float3(1.0,1.0,1.0)).rgb, Opacity);
}

float4 PSSpecular_NoTextureShadowMap(PS_INPUT_NOTEXTURE input) : SV_Target
{
	float3 norm = normalize(input.NormWS);
	
    // Calculate LightTexCoords by using Light ViewProjectionMatrix
    // and dividing by w
    int Split = GetSplit(input.SliceDepth);
    float4 PosLight = mul(float4(input.PosInWorld, 1), SplitMatrices[Split]);
    float2 LightTexCoord = (PosLight.xy / PosLight.w) * float2(0.5, -0.5) + 0.5;
    	
	float3 ShadowContrib = CalcShadowContribution(input.PosInWorld, LightTexCoord, Split);	
    return float4(CalculateSpecularColor(norm, input.PosInWorld, ShadowContrib, float3(1.0,1.0,1.0)).rgb, Opacity);
}

//--------------------------------------------------------------------------------------
technique10 RenderDiffuse
{
    pass P0
    {
		// SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSDiffuse()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderDiffuse_Skin
{
    pass P0
    {
		// SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS_Skin()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSDiffuse()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderDiffuseShadowMap
{
    pass P0
    {
		// SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSDiffuseShadowMap()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderDiffuse_SkinShadowMap
{
    pass P0
    {
		// SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS_Skin()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSDiffuseShadowMap()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderDiffuseLightMap
{
    pass P0
    {
		// SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS_LightMap()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSDiffuseLightMap()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderSHDiffuse
{
    pass P0
    {
		// SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSSHDiffuse()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderSHDiffuse_Skin
{
    pass P0
    {
		// SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS_Skin()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSSHDiffuse()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderSpecular
{
    pass P0
    {
		// SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSSpecular()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderSpecular_Skin
{
    pass P0
    {
		// SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS_Skin()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSSpecular()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderSpecularShadowMap
{
    pass P0
    {
		// SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSSpecularShadowMap()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderSpecular_SkinShadowMap
{
    pass P0
    {
		// SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS_Skin()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSSpecularShadowMap()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderSpecularLightMap
{
    pass P0
    {
		// SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS_LightMap()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSSpecularLightMap()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderDiffuse_NoTexture
{
    pass P0
    {
		//SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS_NoTexture()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSDiffuse_NoTexture()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderDiffuse_NoTexture_Skin
{
    pass P0
    {
		//SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS_NoTexture_Skin()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSDiffuse_NoTexture()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderDiffuse_NoTextureShadowMap
{
    pass P0
    {
		//SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS_NoTexture()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSDiffuse_NoTextureShadowMap()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderDiffuse_NoTexture_SkinShadowMap
{
    pass P0
    {
		//SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS_NoTexture_Skin()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSDiffuse_NoTextureShadowMap()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderSHDiffuse_NoTexture
{
    pass P0
    {
		//SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS_NoTexture()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSSHDiffuse_NoTexture()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderSHDiffuse_NoTexture_Skin
{
    pass P0
    {
		//SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS_NoTexture_Skin()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSSHDiffuse_NoTexture()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderSpecular_NoTexture
{
    pass P0
    {
		//SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS_NoTexture()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSSpecular_NoTexture()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderSpecular_NoTexture_Skin
{
    pass P0
    {
		//SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS_NoTexture_Skin()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSSpecular_NoTexture()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderSpecular_NoTextureShadowMap
{
    pass P0
    {
		//SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS_NoTexture()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSSpecular_NoTextureShadowMap()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderSpecular_NoTexture_SkinShadowMap
{
    pass P0
    {
		//SetRasterizerState(DisableCulling);   
        SetVertexShader(CompileShader(vs_4_0, VS_NoTexture_Skin()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSSpecular_NoTextureShadowMap()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}