// Global objects
//-----------------------------------------
SamplerState samLinear
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
    AddressW = Clamp;
};

SamplerState samPoint
{
    Filter = MIN_MAG_MIP_POINT;
    AddressU = Clamp;
    AddressV = Clamp;
    AddressW = Clamp;
};

RasterizerState cullBack
{
	CullMode = Back;
};

RasterizerState cullFront
{
	CullMode = Front;
};

BlendState SrcAlphaBlendingAdd
{
    BlendEnable[0] = TRUE;
    SrcBlend = SRC_ALPHA;
    DestBlend = INV_SRC_ALPHA;
    BlendOp = ADD;
};

BlendState WriteAlpha
{
    BlendEnable[0] = FALSE;
};


matrix Projection, InvProjection;
matrix View, InvView;
matrix World, InvWorld;
matrix WorldViewProjection;

float Alpha;
//-----------------------------------------



// Box rendering
//-----------------------------------------
struct PS_INPUT_BOX
{
    float4 Pos : SV_POSITION;
    float3 Tex : TEXCOORD0;
};

PS_INPUT_BOX VSBox(float4 pos : POSITION)
{
    PS_INPUT_BOX output;
								// Perspective transform, do not project object coords, move box to the center
    float4 posInView = mul(float4(pos.xyz - 0.5f, 1), mul(World, View));
    output.Pos = mul(posInView, Projection);
    
    output.Tex.xyz = pos.xyz;
    
    return output;
}

float4 PSBox(PS_INPUT_BOX input) : SV_Target
{
    return float4(input.Tex.xyz, input.Pos.z);
}

//-----------------------------------------
// Near clipping plane
//-----------------------------------------
struct PS_INPUT_NEAR_CLIP
{
    float4 Pos : SV_POSITION;
    float3 Tex : TEXCOORD0;
};

PS_INPUT_NEAR_CLIP VSNearClip(float4 pos : POSITION)
{
    PS_INPUT_NEAR_CLIP output;
    output.Pos = pos;

    float4 bindPos = mul(pos, InvProjection);
    bindPos = mul(bindPos, InvView);
    bindPos = mul(bindPos, InvWorld);
    
    output.Tex.xyz = bindPos.xyz / bindPos.w + 0.5f;

    return output;
}

float4 PSNearClip(PS_INPUT_NEAR_CLIP input) : SV_Target
{
    return float4(input.Tex.xyz, 0);
}
//-----------------------------------------
// Ray-casting
//-----------------------------------------
Texture3D txObject;
Texture1D txTransfer;
Texture2D txFront;
Texture2D txBack;
Texture2D txDepth;
Texture2D txWaterDistortion;
TextureCube txEnvironmentMap;
float invWidth, invHeight, invDepth;

struct PS_INPUT_CAST
{
    float4 Pos : SV_POSITION;
    float2 Tex : TEXCOORD0;
};

PS_INPUT_CAST VSCast(float4 pos : POSITION)
{
    PS_INPUT_CAST output;
    output.Pos = pos;
    output.Tex = pos / 2 + 0.5;  // Convert [-1 1] to [0 1]
    output.Tex.y = 1.0f - output.Tex.y;
    return output;
}

float4 PSCast(PS_INPUT_CAST input) : SV_Target
{  
    float4 vecBack;
    vecBack = txBack.Sample(samPoint, input.Tex.xy);
    
    if(vecBack.a == 0.0f)
		discard;
		
    float4 vecFront;
    vecFront = txFront.Sample(samPoint, input.Tex.xy);
		
	float depthPixel = txDepth.Sample(samPoint, input.Tex.xy).x;
    float lenScale = min((depthPixel - vecFront.w) / (vecBack.w - vecFront.w), 1);
    if(lenScale < 0)
		discard;
	
    float3 vecDirAndLen = vecBack - vecFront;
    float3 vecDir = normalize(vecDirAndLen);
		
	// Sample twice per voxel
    vecDir.x *= 0.5f * invWidth;
    vecDir.y *= 0.5f * invHeight;
    vecDir.z *= 0.5f * invDepth;
    vecDirAndLen.x = vecDirAndLen.x / vecDir.x;
    vecDirAndLen.y = vecDirAndLen.y / vecDir.y;
    vecDirAndLen.z = vecDirAndLen.z / vecDir.z;
    float steps = max(ceil(min(min(vecDirAndLen.x, vecDirAndLen.y), vecDirAndLen.z)), 1) * lenScale;
    float depthInc = (vecBack.w - vecFront.w) / steps;
	
    float4 finalColor = float4(0, 0, 0, 0);
    float3 v = vecFront;

    for(int i = 0 ; i < steps ; i++)
    {
		v += vecDir;
			
		float4 color = txTransfer.SampleLevel(samLinear, txObject.SampleLevel(samLinear, v, 0).r, 0);
		
		finalColor.rgb += color.rgb * color.a * (1.0f - finalColor.a);
		finalColor.a += color.a * (1.0f - finalColor.a);
		if(finalColor.a > 0.99) 
			break;
	}
	
    return float4(finalColor.rgb, finalColor.a * Alpha);
}

float4 PSCastAlphaInModel(PS_INPUT_CAST input) : SV_Target
{  
    float4 vecBack;
    vecBack = txBack.Sample(samPoint, input.Tex.xy);
    
    if(vecBack.a == 0.0f)
		discard;
		
    float4 vecFront;
    vecFront = txFront.Sample(samPoint, input.Tex.xy);
		
	float depthPixel = txDepth.Sample(samPoint, input.Tex.xy).x;
    float lenScale = min((depthPixel - vecFront.w) / (vecBack.w - vecFront.w), 1);
    if(lenScale < 0)
		discard;
	
    float3 vecDirAndLen = vecBack - vecFront;
    float3 vecDir = normalize(vecDirAndLen);
		
	// Sample twice per voxel
    vecDir.x *= 0.5f * invWidth;
    vecDir.y *= 0.5f * invHeight;
    vecDir.z *= 0.5f * invDepth;
    vecDirAndLen.x = vecDirAndLen.x / vecDir.x;
    vecDirAndLen.y = vecDirAndLen.y / vecDir.y;
    vecDirAndLen.z = vecDirAndLen.z / vecDir.z;
    float steps = max(ceil(min(min(vecDirAndLen.x, vecDirAndLen.y), vecDirAndLen.z)), 1) * lenScale;
    float depthInc = (vecBack.w - vecFront.w) / steps;
	
    float4 finalColor = float4(0, 0, 0, 0);
    float3 v = vecFront;

    for(int i = 0 ; i < steps ; i++)
    {
		v += vecDir;
			
		float2 obj = txObject.SampleLevel(samLinear, v, 0).rg;
		float4 color = float4(txTransfer.SampleLevel(samLinear, obj.r, 0).rgb, obj.g * 0.45f);
		
		finalColor.rgb += color.rgb * color.a * (1.0f - finalColor.a);
		finalColor.a += color.a * (1.0f - finalColor.a);
		if(finalColor.a > 0.99) 
			break;
	}
		
    return float4(finalColor.rgb, finalColor.a * Alpha);
}

float3 CalcGradientLevelSet(float3 pos)
{
	float3 leftAddr   = (float3(pos.x - invWidth, pos.y, pos.z));
	float3 rightAddr  = (float3(pos.x + invWidth, pos.y, pos.z));
	float3 bottomAddr = (float3(pos.x, pos.y - invHeight, pos.z));
	float3 topAddr    = (float3(pos.x, pos.y + invHeight, pos.z));
	float3 downAddr   = (float3(pos.x, pos.y, pos.z - invDepth));
	float3 upAddr     = (float3(pos.x, pos.y, pos.z + invDepth));

	float left   = txObject.SampleLevel(samLinear, leftAddr, 0).a;
    float right  = txObject.SampleLevel(samLinear, rightAddr, 0).a;
    float bottom = txObject.SampleLevel(samLinear, bottomAddr, 0).a;
    float top    = txObject.SampleLevel(samLinear, topAddr, 0).a;
    float down   = txObject.SampleLevel(samLinear, downAddr, 0).a;
    float up     = txObject.SampleLevel(samLinear, upAddr, 0).a;
	return normalize(float3(right - left, top - bottom, up - down));
}

float4 PSWaterLevelSetCast(PS_INPUT_CAST input) : SV_Target
{  
    float4 vecBack;
    vecBack = txBack.Sample(samPoint, input.Tex.xy);
    
    if(vecBack.a == 0.0f)
		discard;
		
	float4 vecFront;
    vecFront = txFront.Sample(samPoint, input.Tex.xy);
	
    float3 vecDirAndLen = vecBack - vecFront;
    float3 vecDir = normalize(vecDirAndLen);

    float depthPixel = txDepth.Sample(samPoint, input.Tex.xy).x;
    float lenScale = min((depthPixel - vecFront.w) / (vecBack.w - vecFront.w), 1);
    if(lenScale < 0)
		discard;
	
	// Sample once per voxel
	float3 vecDirScaled;
    vecDirScaled.x = vecDir.x * invWidth;
    vecDirScaled.y = vecDir.y * invHeight;
    vecDirScaled.z = vecDir.z * invDepth;
    
    vecDirAndLen.x = vecDirAndLen.x / vecDirScaled.x;
    vecDirAndLen.y = vecDirAndLen.y / vecDirScaled.y;
    vecDirAndLen.z = vecDirAndLen.z / vecDirScaled.z;
    int steps = max(ceil(min(min(vecDirAndLen.x, vecDirAndLen.y), vecDirAndLen.z)), 1) * lenScale;

    float3 v = vecFront;
    float4 color;

    for(int i = 0 ; i < steps ; i++)
    {
		v += vecDirScaled;
		
		color = txObject.SampleLevel(samLinear, v, 0);
		if(color.r < -9)	// whoops, this is an obstacle
			discard;
			
		if(color.a < 0)  // We're in the water
		{
			vecDirScaled *= 0.1f;	// Go back a bit and do better sampling
			float4 color2 = color;
			while(color2.a < 0)
			{
				color = color2;
				v -= vecDirScaled;
				color2 = txObject.SampleLevel(samLinear, v, 0);
			}
			v += vecDirScaled;
				
			float3 fresnelValues = float3(5.0, 0.97963, 0.02037);
			float3 normal = CalcGradientLevelSet(v);
			
			float NdotI = max(dot(-vecDir, normal), 0);
			
			float fresnelTerm = fresnelValues.z + pow(1.0f - NdotI, fresnelValues.x) * fresnelValues.y;


			float3 reflectedView = reflect(vecDir, normal);
			float3 refractedView = refract(vecDir, normal, 1.33f) * vecFront.w;
			
			// Refraction is arbitrary
			float3 refractedColor = txWaterDistortion.SampleLevel(samLinear, input.Tex.xy + refractedView.xy * 0.005f, 0).rgb;
			
			float3 reflectedColor = txEnvironmentMap.SampleLevel(samLinear, reflectedView * 2 -1, 0).rgb;
			// We want more reflection
			return float4(lerp(refractedColor, reflectedColor, 0.35f + max(0.15f, fresnelTerm - 0.02f)), 1.0f);
		}
	}
	discard;
    return float4(0, 0, 0, 0);
}


//-----------------------------------------

technique10 RenderBox
{
    pass P0
    {
        SetRasterizerState(cullBack);
        SetVertexShader(CompileShader(vs_4_0, VSBox()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSBox()));
        SetBlendState(WriteAlpha, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
    
    pass P1
    {
		SetRasterizerState(cullFront);
        SetVertexShader(CompileShader(vs_4_0, VSBox()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSBox()));
        SetBlendState(WriteAlpha, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderNearClip
{
    pass P0
    {
        SetRasterizerState(cullBack);
        SetVertexShader(CompileShader(vs_4_0, VSNearClip()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSNearClip()));
        SetBlendState(WriteAlpha, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderCast
{
    pass P0
    {
        SetRasterizerState(cullBack);
        SetVertexShader(CompileShader(vs_4_0, VSCast()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSCast()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderCastAlphaInModel
{
    pass P0
    {
        SetRasterizerState(cullBack);
        SetVertexShader(CompileShader(vs_4_0, VSCast()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSCastAlphaInModel()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderWaterLevelSetCast
{
    pass P0
    {
        SetRasterizerState(cullBack);
        SetVertexShader(CompileShader(vs_4_0, VSCast()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSWaterLevelSetCast()));
        SetBlendState(SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}